# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''Web页面测试桩代理
'''

import time

# XPath库 Android 2.3以下不支持XPath
# http://www.llamalab.com/js/xpath/XPath.js #该库不支持前后括号情况
# http://coderepos.org/share/wiki/JavaScript-XPath

xpath_lib_str = '''
/*  JavaScript-XPath 0.1.12
 *  (c) 2007 Cybozu Labs, Inc.
 *
 *  JavaScript-XPath is freely distributable under the terms of an MIT-style license.
 *  For details, see the JavaScript-XPath web site: http://coderepos.org/share/wiki/JavaScript-XPath
 *
 */
window.jsxpath={"exportInstaller": true};
(function() {
    var ca = void(0);
    var da = {
        targetFrame: ca,
        exportInstaller: false,
        useNative: true,
        useInnerText: true
    };
    var ea;
    if (window.jsxpath) {
        ea = window.jsxpath;
    } else {
        var fa = document.getElementsByTagName('script');
        var ga = fa[fa.length - 1];
        var ha = ga.src;
        ea = {};
        var ia = ha.match(/\?(.*)$/);
        if (ia) {
            var ja = ia[1].split('&');
            for (var i = 0,
            l = ja.length; i < l; i++) {
                var ka = ja[i];
                var la = ka.split('=');
                var ma = la[0];
                var na = la[1];
                if (na == ca) {
                    na == true;
                } else if (na == 'false' || /^-?\d+$/.test(na)) {
                    na = eval(na);
                }
                ea[ma] = na;
            }
        }
    }
    for (var n in da) {
        if (! (n in ea)) ea[n] = da[n];
    }
    ea.hasNative = !!(document.implementation && document.implementation.hasFeature && document.implementation.hasFeature("XPath", null));
    if (ea.hasNative && ea.useNative && !ea.exportInstaller) {
        return;
    }
    var oa;
    var pa;
    var qa;
    var ra;
    var sa;
    var ta;
    var va;
    var wa;
    var xa;
    var ya;
    var za;
    var Aa;
    var Ba;
    var Ca;
    var Da = new
    function() {
        var ua = navigator.userAgent;
        if (RegExp == ca) {
            if (ua.indexOf("Opera") >= 0) {
                this.opera = true;
            } else if (ua.indexOf("Netscape") >= 0) {
                this.netscape = true;
            } else if (ua.indexOf("Mozilla/") == 0) {
                this.mozilla = true;
            } else {
                this.unknown = true;
            }
            if (ua.indexOf("Gecko/") >= 0) {
                this.gecko = true;
            }
            if (ua.indexOf("Win") >= 0) {
                this.windows = true;
            } else if (ua.indexOf("Mac") >= 0) {
                this.mac = true;
            } else if (ua.indexOf("Linux") >= 0) {
                this.linux = true;
            } else if (ua.indexOf("BSD") >= 0) {
                this.bsd = true;
            } else if (ua.indexOf("SunOS") >= 0) {
                this.sunos = true;
            }
        } else {
            if (/AppleWebKit\/(\d+(?:\.\d+)*)/.test(ua)) {
                this.applewebkit = RegExp.$1;
                if (RegExp.$1.charAt(0) == 4) {
                    this.applewebkit2 = true;
                } else {
                    this.applewebkit3 = true;
                }
            } else if (typeof Components == "object" && (/Gecko\/(\d{8})/.test(ua) || navigator.product == "Gecko" && /^(\d{8})$/.test(navigator.productSub))) {
                this.gecko = RegExp.$1;
            }

            if (typeof(opera) == "object" && typeof(opera.version) == "function") {
                this.opera = opera.version();
                this['opera' + this.opera[0] + this.opera[2]] = true;
            } else if (typeof opera == "object" && (/Opera[\/ ](\d+\.\d+)/.test(ua))) {
                this.opera = RegExp.$1;
            } else if (this.ie) {} else if (/Safari\/(\d+(?:\.\d+)*)/.test(ua)) {
                this.safari = RegExp.$1;
            } else if (/NetFront\/(\d+(?:\.\d+)*)/.test(ua)) {
                this.netfront = RegExp.$1;
            } else if (/Konqueror\/(\d+(?:\.\d+)*)/.test(ua)) {
                this.konqueror = RegExp.$1;
            } else if (ua.indexOf("(compatible;") < 0 && (/^Mozilla\/(\d+\.\d+)/.test(ua))) {
                this.mozilla = RegExp.$1;
                if (/\([^(]*rv:(\d+(?:\.\d+)*).*?\)/.test(ua)) this.mozillarv = RegExp.$1;
                if (/Firefox\/(\d+(?:\.\d+)*)/.test(ua)) {
                    this.firefox = RegExp.$1;
                } else if (/Netscape\d?\/(\d+(?:\.\d+)*)/.test(ua)) {
                    this.netscape = RegExp.$1;
                }
            } else {
                this.unknown = true;
            }
            if (ua.indexOf("Win 9x 4.90") >= 0) {
                this.windows = "ME";
            } else if (/Win(?:dows)? ?(NT ?(\d+\.\d+)?|\d+|ME|Vista|XP)/.test(ua)) {
                this.windows = RegExp.$1;
                if (RegExp.$2) {
                    this.winnt = RegExp.$2;
                } else switch (RegExp.$1) {
                case "2000":
                    this.winnt = "5.0";
                    break;
                case "XP":
                    this.winnt = "5.1";
                    break;
                case "Vista":
                    this.winnt = "6.0";
                    break;
                }
            } else if (ua.indexOf("Mac") >= 0) {
                this.mac = true;
            } else if (ua.indexOf("Linux") >= 0) {
                this.linux = true;
            } else if (/(\w*BSD)/.test(ua)) {
                this.bsd = RegExp.$1;
            } else if (ua.indexOf("SunOS") >= 0) {
                this.sunos = true;
            }
        }
    };
    var Ea = function(Fa) {
        var Ga = Ea.prototype;
        var Ha = Fa.match(Ga.regs.token);
        for (var i = 0,
        l = Ha.length; i < l; i++) {
            if (Ga.regs.strip.test(Ha[i])) {
                Ha.splice(i, 1);
            }
        }
        for (var n in Ga) Ha[n] = Ga[n];
        Ha.index = 0;
        return Ha;
    };
    Ea.prototype.regs = {
        token: /\$?(?:(?![0-9-])[\w-]+:)?(?![0-9-])[\w-]+|\/\/|\.\.|::|\d+(?:\.\d*)?|\.\d+|"[^"]*"|'[^']*'|[!<>]=|(?![0-9-])[\w-]+:\*|\s+|./g,
        strip: /^\s/
    };
    Ea.prototype.peek = function(i) {
        return this[this.index + (i || 0)];
    };
    Ea.prototype.next = function() {
        return this[this.index++];
    };
    Ea.prototype.back = function() {
        this.index--;
    };
    Ea.prototype.empty = function() {
        return this.length <= this.index;
    };
    var Ia = function(Ja, Ka, La) {
        this.node = Ja;
        this.position = Ka || 1;
        this.last = La || 1;
    };
    var Ma = function() {};
    Ma.prototype.number = function(Na) {
        var Oa = this.evaluate(Na);
        if (Oa.isNodeSet) return Oa.number();
        return + Oa;
    };
    Ma.prototype.string = function(Pa) {
        var Qa = this.evaluate(Pa);
        if (Qa.isNodeSet) return Qa.string();
        return '' + Qa;
    };
    Ma.prototype.bool = function(Ra) {
        var Sa = this.evaluate(Ra);
        if (Sa.isNodeSet) return Sa.bool();
        return !! Sa;
    };
    var Ta = function() {};
    Ta.parsePredicates = function(Ua, Va) {
        while (Ua.peek() == '[') {
            Ua.next();
            if (Ua.empty()) {
                throw Error('missing predicate expr');
            }
            var Wa = oa.parse(Ua);
            Va.predicate(Wa);
            if (Ua.empty()) {
                throw Error('unclosed predicate expr');
            }
            if (Ua.next() != ']') {
                Ua.back();
                throw Error('bad token: ' + Ua.next());
            }
        }
    };
    Ta.prototype = new Ma();
    Ta.prototype.evaluatePredicates = function(Xa, Ya) {
        var Za, predicate, nodes, node, Xa, position, reverse;
        reverse = this.reverse;
        Za = this.predicates;
        Xa.sort();
        for (var i = Ya || 0,
        l0 = Za.length; i < l0; i++) {
            predicate = Za[i];
            var $a = [];
            var ab = Xa.list();
            for (var j = 0,
            l1 = ab.length; j < l1; j++) {
                position = reverse ? (l1 - j) : (j + 1);
                exrs = predicate.evaluate(new Ia(ab[j], position, l1));
                switch (typeof exrs) {
                case 'number':
                    exrs = (position == exrs);
                    break;
                case 'string':
                    exrs = !!exrs;
                    break;
                case 'object':
                    exrs = exrs.bool();
                    break;
                }
                if (!exrs) {
                    $a.push(j);
                }
            }
            for (var j = $a.length - 1,
            l1 = 0; j >= l1; j--) {
                Xa.del($a[j]);
            }
        }
        return Xa;
    };
    if (!window.BinaryExpr && window.defaultConfig) window.BinaryExpr = null;
    oa = function(op, bb, cb, db) {
        this.op = op;
        this.left = bb;
        this.right = cb;
        this.datatype = oa.ops[op][2];
        this.needContextPosition = bb.needContextPosition || cb.needContextPosition;
        this.needContextNode = bb.needContextNode || cb.needContextNode;
        if (this.op == '=') {
            if (!cb.needContextNode && !cb.needContextPosition && cb.datatype != 'nodeset' && cb.datatype != 'void' && bb.quickAttr) {
                this.quickAttr = true;
                this.attrName = bb.attrName;
                this.attrValueExpr = cb;
            } else if (!bb.needContextNode && !bb.needContextPosition && bb.datatype != 'nodeset' && bb.datatype != 'void' && cb.quickAttr) {
                this.quickAttr = true;
                this.attrName = cb.attrName;
                this.attrValueExpr = bb;
            }
        }
    };
    oa.compare = function(op, eb, fb, gb, hb) {
        var ib, lnodes, rnodes, nodes, nodeset, primitive;
        fb = fb.evaluate(hb);
        gb = gb.evaluate(hb);
        if (fb.isNodeSet && gb.isNodeSet) {
            lnodes = fb.list();
            rnodes = gb.list();
            for (var i = 0,
            l0 = lnodes.length; i < l0; i++) for (var j = 0,
            l1 = rnodes.length; j < l1; j++) if (eb(wa.to('string', lnodes[i]), wa.to('string', rnodes[j]))) return true;
            return false;
        }
        if (fb.isNodeSet || gb.isNodeSet) {
            if (fb.isNodeSet) nodeset = fb,
            primitive = gb;
            else nodeset = gb,
            primitive = fb;
            nodes = nodeset.list();
            ib = typeof primitive;
            for (var i = 0,
            l = nodes.length; i < l; i++) {
                if (eb(wa.to(ib, nodes[i]), primitive)) return true;
            }
            return false;
        }
        if (op == '=' || op == '!=') {
            if (typeof fb == 'boolean' || typeof gb == 'boolean') {
                return eb( !! fb, !!gb);
            }
            if (typeof fb == 'number' || typeof gb == 'number') {
                return eb( + fb, +gb);
            }
            return eb(fb, gb);
        }
        return eb( + fb, +gb);
    };
    oa.ops = {
        'div': [6,
        function(jb, kb, lb) {
            return jb.number(lb) / kb.number(lb);
        },
        'number'],
        'mod': [6,
        function(mb, nb, ob) {
            return mb.number(ob) % nb.number(ob);
        },
        'number'],
        '*': [6,
        function(pb, qb, rb) {
            return pb.number(rb) * qb.number(rb);
        },
        'number'],
        '+': [5,
        function(sb, tb, ub) {
            return sb.number(ub) + tb.number(ub);
        },
        'number'],
        '-': [5,
        function(vb, wb, xb) {
            return vb.number(xb) - wb.number(xb);
        },
        'number'],
        '<': [4,
        function(yb, zb, Ab) {
            return oa.compare('<',
            function(a, b) {
                return a < b
            },
            yb, zb, Ab);
        },
        'boolean'],
        '>': [4,
        function(Bb, Cb, Db) {
            return oa.compare('>',
            function(a, b) {
                return a > b
            },
            Bb, Cb, Db);
        },
        'boolean'],
        '<=': [4,
        function(Eb, Fb, Gb) {
            return oa.compare('<=',
            function(a, b) {
                return a <= b
            },
            Eb, Fb, Gb);
        },
        'boolean'],
        '>=': [4,
        function(Hb, Ib, Jb) {
            return oa.compare('>=',
            function(a, b) {
                return a >= b
            },
            Hb, Ib, Jb);
        },
        'boolean'],
        '=': [3,
        function(Kb, Lb, Mb) {
            return oa.compare('=',
            function(a, b) {
                return a == b
            },
            Kb, Lb, Mb);
        },
        'boolean'],
        '!=': [3,
        function(Nb, Ob, Pb) {
            return oa.compare('!=',
            function(a, b) {
                return a != b
            },
            Nb, Ob, Pb);
        },
        'boolean'],
        'and': [2,
        function(Qb, Rb, Sb) {
            return Qb.bool(Sb) && Rb.bool(Sb);
        },
        'boolean'],
        'or': [1,
        function(Tb, Ub, Vb) {
            return Tb.bool(Vb) || Ub.bool(Vb);
        },
        'boolean']
    };
    oa.parse = function(Wb) {
        var op, precedence, info, expr, stack = [],
        index = Wb.index;
        while (true) {
            if (Wb.empty()) {
                throw Error('missing right expression');
            }
            expr = Aa.parse(Wb);
            op = Wb.next();
            if (!op) {
                break;
            }
            info = this.ops[op];
            precedence = info && info[0];
            if (!precedence) {
                Wb.back();
                break;
            }
            while (stack.length && precedence <= this.ops[stack[stack.length - 1]][0]) {
                expr = new oa(stack.pop(), stack.pop(), expr);
            }
            stack.push(expr, op);
        }
        while (stack.length) {
            expr = new oa(stack.pop(), stack.pop(), expr);
        }
        return expr;
    };
    oa.prototype = new Ma();
    oa.prototype.evaluate = function(Xb) {
        return oa.ops[this.op][1](this.left, this.right, Xb);
    };
    oa.prototype.show = function(Yb) {
        Yb = Yb || '';
        var t = '';
        t += Yb + 'binary: ' + this.op + '\\n';
        Yb += '    ';
        t += this.left.show(Yb);
        t += this.right.show(Yb);
        return t;
    };
    if (!window.UnaryExpr && window.defaultConfig) window.UnaryExpr = null;
    Aa = function(op, Zb) {
        this.op = op;
        this.expr = Zb;
        this.needContextPosition = Zb.needContextPosition;
        this.needContextNode = Zb.needContextNode;
    };
    Aa.ops = {
        '-': 1
    };
    Aa.parse = function($b) {
        var ac;
        if (this.ops[$b.peek()]) return new Aa($b.next(), Aa.parse($b));
        else return Ba.parse($b);
    };
    Aa.prototype = new Ma();
    Aa.prototype.datatype = 'number';
    Aa.prototype.evaluate = function(bc) {
        return - this.expr.number(bc);
    };
    Aa.prototype.show = function(cc) {
        cc = cc || '';
        var t = '';
        t += cc + 'unary: ' + this.op + '\\n';
        cc += '    ';
        t += this.expr.show(cc);
        return t;
    };
    if (!window.UnionExpr && window.defaultConfig) window.UnionExpr = null;
    Ba = function() {
        this.paths = [];
    };
    Ba.ops = {
        '|': 1
    };
    Ba.parse = function(dc) {
        var ec, expr;
        expr = ya.parse(dc);
        if (!this.ops[dc.peek()]) return expr;
        ec = new Ba();
        ec.path(expr);
        while (true) {
            if (!this.ops[dc.next()]) break;
            if (dc.empty()) {
                throw Error('missing next union location path');
            }
            ec.path(ya.parse(dc));
        }
        dc.back();
        return ec;
    };
    Ba.prototype = new Ma();
    Ba.prototype.datatype = 'nodeset';
    Ba.prototype.evaluate = function(fc) {
        var gc = this.paths;
        var hc = new ta();
        for (var i = 0,
        l = gc.length; i < l; i++) {
            var ic = gc[i].evaluate(fc);
            if (!ic.isNodeSet) throw Error('PathExpr must be nodeset');
            hc.merge(ic);
        }
        return hc;
    };
    Ba.prototype.path = function(jc) {
        this.paths.push(jc);
        if (jc.needContextPosition) {
            this.needContextPosition = true;
        }
        if (jc.needContextNode) {
            this.needContextNode = true;
        }
    }
    Ba.prototype.show = function(kc) {
        kc = kc || '';
        var t = '';
        t += kc + 'union:' + '\\n';
        kc += '    ';
        for (var i = 0; i < this.paths.length; i++) {
            t += this.paths[i].show(kc);
        }
        return t;
    };
    if (!window.PathExpr && window.defaultConfig) window.PathExpr = null;
    ya = function(lc) {
        this.filter = lc;
        this.steps = [];
        this.datatype = lc.datatype;
        this.needContextPosition = lc.needContextPosition;
        this.needContextNode = lc.needContextNode;
    };
    ya.ops = {
        '//': 1,
        '/': 1
    };
    ya.parse = function(mc) {
        var op, expr, path, token;
        if (this.ops[mc.peek()]) {
            op = mc.next();
            token = mc.peek();
            if (op == '/' && (mc.empty() || (token != '.' && token != '..' && token != '@' && token != '*' && !/(?![0-9])[\w]/.test(token)))) {
                return pa.root();
            }
            path = new ya(pa.root());
            if (mc.empty()) {
                throw Error('missing next location step');
            }
            expr = za.parse(mc);
            path.step(op, expr);
        } else {
            expr = pa.parse(mc);
            if (!expr) {
                expr = za.parse(mc);
                path = new ya(pa.context());
                path.step('/', expr);
            } else if (!this.ops[mc.peek()]) return expr;
            else path = new ya(expr);
        }
        while (true) {
            if (!this.ops[mc.peek()]) break;
            op = mc.next();
            if (mc.empty()) {
                throw Error('missing next location step');
            }
            path.step(op, za.parse(mc));
        }
        return path;
    };
    ya.prototype = new Ma();
    ya.prototype.evaluate = function(nc) {
        var oc = this.filter.evaluate(nc);
        if (!oc.isNodeSet) throw Exception('Filter nodeset must be nodeset type');
        var pc = this.steps;
        for (var i = 0,
        l0 = pc.length; i < l0 && oc.length; i++) {
            var qc = pc[i][1];
            var rc = qc.reverse;
            var sc = oc.iterator(rc);
            var tc = oc;
            oc = null;
            var uc, next;
            if (!qc.needContextPosition && qc.axis == 'following') {
                for (uc = sc(); next = sc(); uc = next) {
                    if (Da.applewebkit2) {
                        var vc = false;
                        var wc = next;
                        do {
                            if (wc == uc) {
                                vc = true;
                                break;
                            }
                        } while ( wc = wc . parentNode );
                        if (!vc) break;
                    } else {
                        try {
                            if (!uc.contains(next)) break
                        } catch(e) {
                            if (! (next.compareDocumentPosition(uc) & 8)) break
                        }
                    }
                }
                oc = qc.evaluate(new Ia(uc));
            } else if (!qc.needContextPosition && qc.axis == 'preceding') {
                uc = sc();
                oc = qc.evaluate(new Ia(uc));
            } else {
                uc = sc();
                var j = 0;
                oc = qc.evaluate(new Ia(uc), false, tc, j);
                while (uc = sc()) {
                    j++;
                    oc.merge(qc.evaluate(new Ia(uc), false, tc, j));
                }
            }
        }
        return oc;
    };
    ya.prototype.step = function(op, xc) {
        xc.op = op;
        this.steps.push([op, xc]);
        this.quickAttr = false;
        if (this.steps.length == 1) {
            if (op == '/' && xc.axis == 'attribute') {
                var yc = xc.test;
                if (!yc.notOnlyElement && yc.name != '*') {
                    this.quickAttr = true;
                    this.attrName = yc.name;
                }
            }
        }
    };
    ya.prototype.show = function(zc) {
        zc = zc || '';
        var t = '';
        t += zc + 'path:' + '\\n';
        zc += '    ';
        t += zc + 'filter:' + '\\n';
        t += this.filter.show(zc + '    ');
        if (this.steps.length) {
            t += zc + 'steps:' + '\\n';
            zc += '    ';
            for (var i = 0; i < this.steps.length; i++) {
                var Ac = this.steps[i];
                t += zc + 'operator: ' + Ac[0] + '\\n';
                t += Ac[1].show(zc);
            }
        }
        return t;
    };
    if (!window.FilterExpr && window.defaultConfig) window.FilterExpr = null;
    pa = function(Bc) {
        this.primary = Bc;
        this.predicates = [];
        this.datatype = Bc.datatype;
        this.needContextPosition = Bc.needContextPosition;
        this.needContextNode = Bc.needContextNode;
    };
    pa.parse = function(Cc) {
        var Dc, filter, token, ch;
        token = Cc.peek();
        ch = token.charAt(0);
        switch (ch) {
        case '$':
            Dc = Ca.parse(Cc);
            break;
        case '(':
            Cc.next();
            Dc = oa.parse(Cc);
            if (Cc.empty()) {
                throw Error('unclosed "("');
            }
            if (Cc.next() != ')') {
                Cc.back();
                throw Error('bad token: ' + Cc.next());
            }
            break;
        case '"':
        case "'":
            Dc = ra.parse(Cc);
            break;
        default:
            if (!isNaN( + token)) {
                Dc = xa.parse(Cc);
            } else if (va.types[token]) {
                return null;
            } else if (/(?![0-9])[\w]/.test(ch) && Cc.peek(1) == '(') {
                Dc = qa.parse(Cc);
            } else {
                return null;
            }
            break;
        }
        if (Cc.peek() != '[') return Dc;
        filter = new pa(Dc);
        Ta.parsePredicates(Cc, filter);
        return filter;
    };
    pa.root = function() {
        return new qa('root-node');
    };
    pa.context = function() {
        return new qa('context-node');
    };
    pa.prototype = new Ta();
    pa.prototype.evaluate = function(Ec) {
        var Fc = this.primary.evaluate(Ec);
        if (!Fc.isNodeSet) {
            if (this.predicates.length) throw Error('Primary result must be nodeset type ' + 'if filter have predicate expression');
            return Fc;
        }
        return this.evaluatePredicates(Fc);
    };
    pa.prototype.predicate = function(Gc) {
        this.predicates.push(Gc);
    };
    pa.prototype.show = function(Hc) {
        Hc = Hc || '';
        var t = '';
        t += Hc + 'filter: ' + '\\n';
        Hc += '    ';
        t += this.primary.show(Hc);
        if (this.predicates.length) {
            t += Hc + 'predicates: ' + '\\n';
            Hc += '    ';
            for (var i = 0; i < this.predicates.length; i++) {
                t += this.predicates[i].show(Hc);
            }
        }
        return t;
    };
    if (!window.NodeUtil && window.defaultConfig) window.NodeUtil = null;
    wa = {
        to: function(Ic, Jc) {
            var t, type = Jc.nodeType;
            if (type == 1 && ea.useInnerText && !Da.applewebkit2) {
                t = Jc.textContent;
                t = (t == ca || t == null) ? Jc.innerText: t;
                t = (t == ca || t == null) ? '': t;
            }
            if (typeof t != 'string') {
                if (type == 9 || type == 1) {
                    if (type == 9) {
                        Jc = Jc.documentElement;
                    } else {
                        Jc = Jc.firstChild;
                    }
                    for (t = '', stack = [], i = 0; Jc;) {
                        do {
                            if (Jc.nodeType != 1) {
                                t += Jc.nodeValue;
                            }
                            stack[i++] = Jc;
                        } while ( Jc = Jc . firstChild );
                        while (i && !(Jc = stack[--i].nextSibling)) {}
                    }
                } else {
                    t = Jc.nodeValue;
                }
            }
            switch (Ic) {
            case 'number':
                return + t;
            case 'boolean':
                return !! t;
            default:
                return t;
            }
        },
        attrPropMap: {
            name: 'name',
            'class': 'className',
            dir: 'dir',
            id: 'id',
            name: 'name',
            title: 'title'
        },
        attrMatch: function(Kc, Lc, Mc) {
            if (!Lc || Mc == null && Kc.hasAttribute && Kc.hasAttribute(Lc) || Mc != null && Kc.getAttribute && Kc.getAttribute(Lc) == Mc) {
                return true;
            } else {
                return false;
            }
        },
        getDescendantNodes: function(Oc, Kc, Pc, Lc, Mc, Qc, Rc) {
            if (Qc) {
                Qc.delDescendant(Kc, Rc);
            }
            if (Mc && Lc == 'id' && Kc.getElementById) {
                Kc = Kc.getElementById(Mc);
                if (Kc && Oc.match(Kc)) {
                    Pc.push(Kc);
                }
            } else if (Mc && Lc == 'name' && Kc.getElementsByName) {
                var Wc = Kc.getElementsByName(Mc);
                for (var i = 0,
                l = Wc.length; i < l; i++) {
                    Kc = Wc[i];
                    if (Da.opera ? (Kc.name == Mc && Oc.match(Kc)) : Oc.match(Kc)) {
                        Pc.push(Kc);
                    }
                }
            } else if (Mc && Lc == 'class' && Kc.getElementsByClassName) {
                var Wc = Kc.getElementsByClassName(Mc);
                for (var i = 0,
                l = Wc.length; i < l; i++) {
                    Kc = Wc[i];
                    if (Kc.className == Mc && Oc.match(Kc)) {
                        Pc.push(Kc);
                    }
                }
            } else if (Oc.notOnlyElement) { (function(Xc) {
                    var f = arguments.callee;
                    for (var Kc = Xc.firstChild; Kc; Kc = Kc.nextSibling) {
                        if (wa.attrMatch(Kc, Lc, Mc)) {
                            if (Oc.match(Kc.nodeType)) Pc.push(Kc);
                        }
                        f(Kc);
                    }
                })(Kc);
            } else {
                var Tc = Oc.name;
                if (Kc.getElementsByTagName) {
                    var Wc = Kc.getElementsByTagName(Tc);
                    if (Wc) {
                        var i = 0;
                        while (Kc = Wc[i++]) {
                            if (wa.attrMatch(Kc, Lc, Mc)) Pc.push(Kc);
                        }
                    }
                }
            }
            return Pc;
        },
        getChildNodes: function(Yc, Kc, Zc, Lc, Mc) {
            for (var Kc = Kc.firstChild; Kc; Kc = Kc.nextSibling) {
                if (wa.attrMatch(Kc, Lc, Mc)) {
                    if (Yc.match(Kc)) Zc.push(Kc);
                }
            }
            return Zc;
        }
    };
    
    if (!window.Step && window.defaultConfig) window.Step = null;
    za = function(gd, hd) {
        this.axis = gd;
        this.reverse = za.axises[gd][0];
        this.func = za.axises[gd][1];
        this.test = hd;
        this.predicates = [];
        this._quickAttr = za.axises[gd][2]
    };
    za.axises = {
        ancestor: [true,
        function(jd, kd, ld, _, md, od, pd) {
            while (kd = kd.parentNode) {
                if (od && kd.nodeType == 1) {
                    od.reserveDelByNode(kd, pd, true);
                }
                if (jd.match(kd)) ld.unshift(kd);
            }
            return ld;
        }],
        'ancestor-or-self': [true,
        function(qd, rd, sd, _, td, ud, vd) {
            do {
                if (ud && rd.nodeType == 1) {
                    ud.reserveDelByNode(rd, vd, true);
                }
                if (qd.match(rd)) sd.unshift(rd);
            } while ( rd = rd . parentNode ) return sd;
        }],
        attribute: [false,
        function(wd, xd, yd) {
            var zd = xd.attributes;
            if (zd) {
                if ((wd.notOnlyElement && wd.type == 0) || wd.name == '*') {
                    for (var i = 0,
                    attr; attr = zd[i]; i++) {
                        yd.push(attr);
                    }
                } else {
                    var Bd = zd.getNamedItem(wd.name);
                    if (Bd) {
                        yd.push(Bd);
                    }
                }
            }
            return yd;
        }],
        child: [false, wa.getChildNodes, true],
        descendant: [false, wa.getDescendantNodes, true],
        'descendant-or-self': [false,
        function(wd, xd, yd, Cd, Dd, Ed, Fd) {
            if (wa.attrMatch(xd, Cd, Dd)) {
                if (wd.match(xd)) yd.push(xd);
            }
            return wa.getDescendantNodes(wd, xd, yd, Cd, Dd, Ed, Fd);
        },
        true],
        following: [false,
        function(wd, xd, yd, Gd, Hd) {
            do {
                var Id = xd;
                while (Id = Id.nextSibling) {
                    if (wa.attrMatch(Id, Gd, Hd)) {
                        if (wd.match(Id)) yd.push(Id);
                    }
                    yd = wa.getDescendantNodes(wd, Id, yd, Gd, Hd);
                }
            } while ( xd = xd . parentNode );
            return yd;
        },
        true],
        'following-sibling': [false,
        function(wd, xd, yd, _, Jd, Kd, Ld) {
            while (xd = xd.nextSibling) {
                if (Kd && xd.nodeType == 1) {
                    Kd.reserveDelByNode(xd, Ld);
                }
                if (wd.match(xd)) {
                    yd.push(xd);
                }
            }
            return yd;
        }],
        namespace: [false,
        function(wd, xd, yd) {
            return yd;
        }],
        parent: [false,
        function(wd, xd, yd) {
            if (xd.nodeType == 9) {
                return yd;
            }
            if (xd.nodeType == 2) {
                yd.push(xd.ownerElement);
                return yd;
            }
            var xd = xd.parentNode;
            if (wd.match(xd)) yd.push(xd);
            return yd;
        }],
        preceding: [true,
        function(wd, xd, yd, Md, Nd) {
            var Od = [];
            do {
                Od.unshift(xd);
            } while ( xd = xd . parentNode );
            for (var i = 1,
            l0 = Od.length; i < l0; i++) {
                var Pd = [];
                xd = Od[i];
                while (xd = xd.previousSibling) {
                    Pd.unshift(xd);
                }
                for (var j = 0,
                l1 = Pd.length; j < l1; j++) {
                    xd = Pd[j];
                    if (wa.attrMatch(xd, Md, Nd)) {
                        if (wd.match(xd)) yd.push(xd);
                    }
                    yd = wa.getDescendantNodes(wd, xd, yd, Md, Nd);
                }
            }
            return yd;
        },
        true],
        'preceding-sibling': [true,
        function(wd, xd, yd, _, Qd, Rd, Sd) {
            while (xd = xd.previousSibling) {
                if (Rd && xd.nodeType == 1) {
                    Rd.reserveDelByNode(xd, Sd, true);
                }
                if (wd.match(xd)) {
                    yd.unshift(xd)
                }
            }
            return yd;
        }],
        self: [false,
        function(wd, xd, yd) {
            if (wd.match(xd)) yd.push(xd);
            return yd;
        }]
    };
    za.parse = function(Td) {
        var Ud, test, step, token;
        if (Td.peek() == '.') {
            step = this.self();
            Td.next();
        } else if (Td.peek() == '..') {
            step = this.parent();
            Td.next();
        } else {
            if (Td.peek() == '@') {
                Ud = 'attribute';
                Td.next();
                if (Td.empty()) {
                    throw Error('missing attribute name');
                }
            } else {
                if (Td.peek(1) == '::') {
                    if (!/(?![0-9])[\w]/.test(Td.peek().charAt(0))) {
                        throw Error('bad token: ' + Td.next());
                    }
                    Ud = Td.next();
                    Td.next();
                    if (!this.axises[Ud]) {
                        throw Error('invalid axis: ' + Ud);
                    }
                    if (Td.empty()) {
                        throw Error('missing node name');
                    }
                } else {
                    Ud = 'child';
                }
            }
            token = Td.peek();
            if (!/(?![0-9])[\w]/.test(token.charAt(0))) {
                if (token == '*') {
                    test = sa.parse(Td)
                } else {
                    throw Error('bad token: ' + Td.next());
                }
            } else {
                if (Td.peek(1) == '(') {
                    if (!va.types[token]) {
                        throw Error('invalid node type: ' + token);
                    }
                    test = va.parse(Td)
                } else {
                    test = sa.parse(Td);
                }
            }
            step = new za(Ud, test);
        }
        Ta.parsePredicates(Td, step);
        return step;
    };
    za.self = function() {
        return new za('self', new va('node'));
    };
    za.parent = function() {
        return new za('parent', new va('node'));
    };
    za.prototype = new Ta();
    za.prototype.evaluate = function(Vd, Wd, Xd, Yd) {
        var Zd = Vd.node;
        var $d = false;
        if (!Wd && this.op == '//') {
            if (!this.needContextPosition && this.axis == 'child') {
                if (this.quickAttr) {
                    var ae = this.attrValueExpr ? this.attrValueExpr.string(Vd) : null;
                    var be = wa.getDescendantNodes(this.test, Zd, new ta(), this.attrName, ae, Xd, Yd);
                    be = this.evaluatePredicates(be, 1);
                } else {
                    var be = wa.getDescendantNodes(this.test, Zd, new ta(), null, null, Xd, Yd);
                    be = this.evaluatePredicates(be);
                }
            } else {
                var ce = new za('descendant-or-self', new va('node'));
                var de = ce.evaluate(Vd, false, Xd, Yd).list();
                var be = null;
                ce.op = '/';
                for (var i = 0,
                l = de.length; i < l; i++) {
                    if (!be) {
                        be = this.evaluate(new Ia(de[i]), true);
                    } else {
                        be.merge(this.evaluate(new Ia(de[i]), true));
                    }
                }
                be = be || new ta();
            }
        } else {
            if (this.needContextPosition) {
                Xd = null;
                Yd = null;
            }
            if (this.quickAttr) {
                var ae = this.attrValueExpr ? this.attrValueExpr.string(Vd) : null;
                var be = this.func(this.test, Zd, new ta(), this.attrName, ae, Xd, Yd);
                be = this.evaluatePredicates(be, 1);
            } else {
                var be = this.func(this.test, Zd, new ta(), null, null, Xd, Yd);
                be = this.evaluatePredicates(be);
            }
            if (Xd) {
                Xd.doDel();
            }
        }
        return be;
    };
    za.prototype.predicate = function(ee) {
        this.predicates.push(ee);
        if (ee.needContextPosition || ee.datatype == 'number' || ee.datatype == 'void') {
            this.needContextPosition = true;
        }
        if (this._quickAttr && this.predicates.length == 1 && ee.quickAttr) {
            var fe = ee.attrName;
            this.attrName = fe;
            this.attrValueExpr = ee.attrValueExpr;
            this.quickAttr = true;
        }
    };
    za.prototype.show = function(ge) {
        ge = ge || '';
        var t = '';
        t += ge + 'step: ' + '\\n';
        ge += '    ';
        if (this.axis) t += ge + 'axis: ' + this.axis + '\\n';
        t += this.test.show(ge);
        if (this.predicates.length) {
            t += ge + 'predicates: ' + '\\n';
            ge += '    ';
            for (var i = 0; i < this.predicates.length; i++) {
                t += this.predicates[i].show(ge);
            }
        }
        return t;
    };
    if (!window.NodeType && window.defaultConfig) window.NodeType = null;
    va = function(he, je) {
        this.name = he;
        this.literal = je;
        switch (he) {
        case 'comment':
            this.type = 8;
            break;
        case 'text':
            this.type = 3;
            break;
        case 'processing-instruction':
            this.type = 7;
            break;
        case 'node':
            this.type = 0;
            break;
        }
    };
    va.types = {
        'comment': 1,
        'text': 1,
        'processing-instruction': 1,
        'node': 1
    };
    va.parse = function(ke) {
        var le, literal, ch;
        le = ke.next();
        ke.next();
        if (ke.empty()) {
            throw Error('bad nodetype');
        }
        ch = ke.peek().charAt(0);
        if (ch == '"' || ch == "'") {
            literal = ra.parse(ke);
        }
        if (ke.empty()) {
            throw Error('bad nodetype');
        }
        if (ke.next() != ')') {
            ke.back();
            throw Error('bad token ' + ke.next());
        }
        return new va(le, literal);
    };
    va.prototype = new Ma();
    va.prototype.notOnlyElement = true;
    va.prototype.match = function(me) {
        return ! this.type || this.type == me.nodeType;
    };
    va.prototype.show = function(ne) {
        ne = ne || '';
        var t = '';
        t += ne + 'nodetype: ' + this.type + '\\n';
        if (this.literal) {
            ne += '    ';
            t += this.literal.show(ne);
        }
        return t;
    };
    if (!window.NameTest && window.defaultConfig) window.NameTest = null;
    sa = function(oe) {
        this.name = oe.toLowerCase();
    };
    sa.parse = function(pe) {
        if (pe.peek() != '*' && pe.peek(1) == ':' && pe.peek(2) == '*') {
            return new sa(pe.next() + pe.next() + pe.next());
        }
        return new sa(pe.next());
    };
    sa.prototype = new Ma();
    sa.prototype.match = function(qe) {
        var re = qe.nodeType;
        if (re == 1 || re == 2) {
            if (this.name == '*' || this.name == qe.nodeName.toLowerCase()) {
                return true;
            }
        }
        return false;
    };
    sa.prototype.show = function(se) {
        se = se || '';
        var t = '';
        t += se + 'nametest: ' + this.name + '\\n';
        return t;
    };
    if (!window.VariableReference && window.defaultConfig) window.VariableReference = null;
    Ca = function(te) {
        this.name = te.substring(1);
    };
    Ca.parse = function(ue) {
        var ve = ue.next();
        if (ve.length < 2) {
            throw Error('unnamed variable reference');
        }
        return new Ca(ve)
    };
    Ca.prototype = new Ma();
    Ca.prototype.datatype = 'void';
    Ca.prototype.show = function(we) {
        we = we || '';
        var t = '';
        t += we + 'variable: ' + this.name + '\\n';
        return t;
    };
    if (!window.Literal && window.defaultConfig) window.Literal = null;
    ra = function(xe) {
        this.text = xe.substring(1, xe.length - 1);
    };
    ra.parse = function(ye) {
        var ze = ye.next();
        if (ze.length < 2) {
            throw Error('unclosed literal string');
        }
        return new ra(ze)
    };
    ra.prototype = new Ma();
    ra.prototype.datatype = 'string';
    ra.prototype.evaluate = function(Ae) {
        return this.text;
    };
    ra.prototype.show = function(Be) {
        Be = Be || '';
        var t = '';
        t += Be + 'literal: ' + this.text + '\\n';
        return t;
    };
    if (!window.Number && window.defaultConfig) window.Number = null;
    xa = function(Ce) {
        this.digit = +Ce;
    };
    xa.parse = function(De) {
        return new xa(De.next());
    };
    xa.prototype = new Ma();
    xa.prototype.datatype = 'number';
    xa.prototype.evaluate = function(Ee) {
        return this.digit;
    };
    xa.prototype.show = function(Fe) {
        Fe = Fe || '';
        var t = '';
        t += Fe + 'number: ' + this.digit + '\\n';
        return t;
    };
    if (!window.FunctionCall && window.defaultConfig) window.FunctionCall = null;
    qa = function(Ge) {
        var He = qa.funcs[Ge];
        if (!He) throw Error(Ge + ' is not a function');
        this.name = Ge;
        this.func = He[0];
        this.args = [];
        this.datatype = He[1];
        if (He[2]) {
            this.needContextPosition = true;
        }
        this.needContextNodeInfo = He[3];
        this.needContextNode = this.needContextNodeInfo[0]
    };
    qa.funcs = {
        'context-node': [function() {
            if (arguments.length != 0) {
                throw Error('Function context-node expects ()');
            }
            var ns;
            ns = new ta();
            ns.push(this.node);
            return ns;
        },
        'nodeset', false, [true]],
        'root-node': [function() {
            if (arguments.length != 0) {
                throw Error('Function root-node expects ()');
            }
            var ns, ctxn;
            ns = new ta();
            ctxn = this.node;
            if (ctxn.nodeType == 9) ns.push(ctxn);
            else ns.push(ctxn.ownerDocument);
            return ns;
        },
        'nodeset', false, []],
        last: [function() {
            if (arguments.length != 0) {
                throw Error('Function last expects ()');
            }
            return this.last;
        },
        'number', true, []],
        position: [function() {
            if (arguments.length != 0) {
                throw Error('Function position expects ()');
            }
            return this.position;
        },
        'number', true, []],
        count: [function(ns) {
            if (arguments.length != 1 || !(ns = ns.evaluate(this)).isNodeSet) {
                throw Error('Function count expects (nodeset)');
            }
            return ns.length;
        },
        'number', false, []],
        id: [function(s) {
            var Ie, ns, i, id, elm, ctxn, doc;
            if (arguments.length != 1) {
                throw Error('Function id expects (object)');
            }
            ctxn = this.node;
            if (ctxn.nodeType == 9) doc = ctxn;
            else doc = ctxn.ownerDocument;
            s = s.string(this);
            Ie = s.split(/\s+/);
            ns = new ta();
            for (i = 0, l = Ie.length; i < l; i++) {
                id = Ie[i];
                elm = doc.getElementById(id);
                if (Da.opera && elm && elm.id != id) {
                    var Je = doc.getElementsByName(id);
                    for (var j = 0,
                    l0 = Je.length; j < l0; j++) {
                        elm = Je[j];
                        if (elm.id == id) {
                            ns.push(elm);
                        }
                    }
                } else {
                    if (elm) ns.push(elm)
                }
            }
            ns.isSorted = false;
            return ns;
        },
        'nodeset', false, []],
        'local-name': [function(ns) {
            var nd;
            switch (arguments.length) {
            case 0:
                nd = this.node;
                break;
            case 1:
                if ((ns = ns.evaluate(this)).isNodeSet) {
                    nd = ns.first();
                    break;
                }
            default:
                throw Error('Function local-name expects (nodeset?)');
                break;
            }
            return '' + nd.nodeName.toLowerCase();
        },
        'string', false, [true, false]],
        name: [function(ns) {
            return qa.funcs['local-name'][0].apply(this, arguments);
        },
        'string', false, [true, false]],
        'namespace-uri': [function(ns) {
            return '';
        },
        'string', false, [true, false]],
        string: [function(s) {
            switch (arguments.length) {
            case 0:
                s = wa.to('string', this.node);
                break;
            case 1:
                s = s.string(this);
                break;
            default:
                throw Error('Function string expects (object?)');
                break;
            }
            return s;
        },
        'string', false, [true, false]],
        concat: [function(s1, s2) {
            if (arguments.length < 2) {
                throw Error('Function concat expects (string, string[, ...])');
            }
            for (var t = '',
            i = 0,
            l = arguments.length; i < l; i++) {
                t += arguments[i].string(this);
            }
            return t;
        },
        'string', false, []],
        'starts-with': [function(s1, s2) {
            if (arguments.length != 2) {
                throw Error('Function starts-with expects (string, string)');
            }
            s1 = s1.string(this);
            s2 = s2.string(this);
            return s1.indexOf(s2) == 0;
        },
        'boolean', false, []],
        contains: [function(s1, s2) {
            if (arguments.length != 2) {
                throw Error('Function contains expects (string, string)');
            }
            s1 = s1.string(this);
            s2 = s2.string(this);
            return s1.indexOf(s2) != -1;
        },
        'boolean', false, []],
        substring: [function(s, n1, n2) {
            var a1, a2;
            s = s.string(this);
            n1 = n1.number(this);
            switch (arguments.length) {
            case 2:
                n2 = s.length - n1 + 1;
                break;
            case 3:
                n2 = n2.number(this);
                break;
            default:
                throw Error('Function substring expects (string, string)');
                break;
            }
            n1 = Math.round(n1);
            n2 = Math.round(n2);
            a1 = n1 - 1;
            a2 = n1 + n2 - 1;
            if (a2 == Infinity) {
                return s.substring(a1 < 0 ? 0 : a1);
            } else {
                return s.substring(a1 < 0 ? 0 : a1, a2)
            }
        },
        'string', false, []],
        'substring-before': [function(s1, s2) {
            var n;
            if (arguments.length != 2) {
                throw Error('Function substring-before expects (string, string)');
            }
            s1 = s1.string(this);
            s2 = s2.string(this);
            n = s1.indexOf(s2);
            if (n == -1) return '';
            return s1.substring(0, n);
        },
        'string', false, []],
        'substring-after': [function(s1, s2) {
            if (arguments.length != 2) {
                throw Error('Function substring-after expects (string, string)');
            }
            s1 = s1.string(this);
            s2 = s2.string(this);
            var n = s1.indexOf(s2);
            if (n == -1) return '';
            return s1.substring(n + s2.length);
        },
        'string', false, []],
        'string-length': [function(s) {
            switch (arguments.length) {
            case 0:
                s = wa.to('string', this.node);
                break;
            case 1:
                s = s.string(this);
                break;
            default:
                throw Error('Function string-length expects (string?)');
                break;
            }
            return s.length;
        },
        'number', false, [true, false]],
        'normalize-space': [function(s) {
            switch (arguments.length) {
            case 0:
                s = wa.to('string', this.node);
                break;
            case 1:
                s = s.string(this);
                break;
            default:
                throw Error('Function normalize-space expects (string?)');
                break;
            }
            return s.replace(/\s+/g, ' ').replace(/^ /, '').replace(/ $/, '');
        },
        'string', false, [true, false]],
        translate: [function(s1, s2, s3) {
            if (arguments.length != 3) {
                throw Error('Function translate expects (string, string, string)');
            }
            s1 = s1.string(this);
            s2 = s2.string(this);
            s3 = s3.string(this);
            var Le = [];
            for (var i = 0,
            l = s2.length; i < l; i++) {
                var ch = s2.charAt(i);
                if (!Le[ch]) Le[ch] = s3.charAt(i) || '';
            }
            for (var t = '',
            i = 0,
            l = s1.length; i < l; i++) {
                var ch = s1.charAt(i);
                var Me = Le[ch];
                t += (Me != ca) ? Me: ch;
            }
            return t;
        },
        'string', false, []],
        'boolean': [function(b) {
            if (arguments.length != 1) {
                throw Error('Function boolean expects (object)');
            }
            return b.bool(this)
        },
        'boolean', false, []],
        not: [function(b) {
            if (arguments.length != 1) {
                throw Error('Function not expects (object)');
            }
            return ! b.bool(this)
        },
        'boolean', false, []],
        'true': [function() {
            if (arguments.length != 0) {
                throw Error('Function true expects ()');
            }
            return true;
        },
        'boolean', false, []],
        'false': [function() {
            if (arguments.length != 0) {
                throw Error('Function false expects ()');
            }
            return false;
        },
        'boolean', false, []],
        lang: [function(s) {
            return false;
        },
        'boolean', false, []],
        number: [function(n) {
            switch (arguments.length) {
            case 0:
                n = wa.to('number', this.node);
                break;
            case 1:
                n = n.number(this);
                break;
            default:
                throw Error('Function number expects (object?)');
                break;
            }
            return n;
        },
        'number', false, [true, false]],
        sum: [function(ns) {
            var Ne, n, i, l;
            if (arguments.length != 1 || !(ns = ns.evaluate(this)).isNodeSet) {
                throw Error('Function sum expects (nodeset)');
            }
            Ne = ns.list();
            n = 0;
            for (i = 0, l = Ne.length; i < l; i++) {
                n += wa.to('number', Ne[i]);
            }
            return n;
        },
        'number', false, []],
        floor: [function(n) {
            if (arguments.length != 1) {
                throw Error('Function floor expects (number)');
            }
            n = n.number(this);
            return Math.floor(n);
        },
        'number', false, []],
        ceiling: [function(n) {
            if (arguments.length != 1) {
                throw Error('Function ceiling expects (number)');
            }
            n = n.number(this);
            return Math.ceil(n);
        },
        'number', false, []],
        round: [function(n) {
            if (arguments.length != 1) {
                throw Error('Function round expects (number)');
            }
            n = n.number(this);
            return Math.round(n);
        },
        'number', false, []]
    };
    qa.parse = function(Oe) {
        var Pe, func = new qa(Oe.next());
        Oe.next();
        while (Oe.peek() != ')') {
            if (Oe.empty()) {
                throw Error('missing function argument list');
            }
            Pe = oa.parse(Oe);
            func.arg(Pe);
            if (Oe.peek() != ',') break;
            Oe.next();
        }
        if (Oe.empty()) {
            throw Error('unclosed function argument list');
        }
        if (Oe.next() != ')') {
            Oe.back();
            throw Error('bad token: ' + Oe.next());
        }
        return func
    };
    qa.prototype = new Ma();
    qa.prototype.evaluate = function(Qe) {
        return this.func.apply(Qe, this.args);
    };
    qa.prototype.arg = function(Re) {
        this.args.push(Re);
        if (Re.needContextPosition) {
            this.needContextPosition = true;
        }
        var Se = this.args;
        if (Re.needContextNode) {
            Se.needContexNode = true;
        }
        this.needContextNode = Se.needContextNode || this.needContextNodeInfo[Se.length];
    };
    qa.prototype.show = function(Te) {
        Te = Te || '';
        var t = '';
        t += Te + 'function: ' + this.name + '\\n';
        Te += '    ';
        if (this.args.length) {
            t += Te + 'arguments: ' + '\\n';
            Te += '    ';
            for (var i = 0; i < this.args.length; i++) {
                t += this.args[i].show(Te);
            }
        }
        return t;
    };
    var Ze = {
        uuid: 1,
        get: function($e) {
            return $e.__ba || ($e.__ba = this.uuid++);
        }
    };

    if (!window.NodeSet && window.defaultConfig) window.NodeSet = null;
    ta = function() {
        this.length = 0;
        this.nodes = [];
        this.seen = {};
        this.idIndexMap = null;
        this.reserveDels = [];
    };
    ta.prototype.isNodeSet = true;
    ta.prototype.isSorted = true;
    ta.prototype.merge = function(af) {
        this.isSorted = false;
        if (af.only) {
            return this.push(af.only);
        }
        if (this.only) {
            var bf = this.only;
            delete this.only;
            this.push(bf);
            this.length--;
        }
        var cf = af.nodes;
        for (var i = 0,
        l = cf.length; i < l; i++) {
            this._add(cf[i]);
        }
    };
    ta.prototype.sort = function() {
        if (this.only) return;
        if (this.sortOff) return;
        if (!this.isSorted) {
            this.isSorted = true;
            this.idIndexMap = null;
            var ef = this.nodes;
            ef.sort(function(a, b) {
                if (a == b) return 0;
                if (a.compareDocumentPosition) {
                    var ff = a.compareDocumentPosition(b);
                    if (ff & 2) return 1;
                    if (ff & 4) return - 1;
                    return 0;
                } else {
                    var gf = a,
                    node2 = b,
                    ancestor1 = a,
                    ancestor2 = b,
                    deep1 = 0,
                    deep2 = 0;
                    while (ancestor1 = ancestor1.parentNode) deep1++;
                    while (ancestor2 = ancestor2.parentNode) deep2++;
                    if (deep1 > deep2) {
                        while (deep1--!=deep2) gf = gf.parentNode;
                        if (gf == node2) return 1;
                    } else if (deep2 > deep1) {
                        while (deep2--!=deep1) node2 = node2.parentNode;
                        if (gf == node2) return - 1;
                    }
                    while ((ancestor1 = gf.parentNode) != (ancestor2 = node2.parentNode)) {
                        gf = ancestor1;
                        node2 = ancestor2;
                    }
                    while (gf = gf.nextSibling) if (gf == node2) return - 1;
                    return 1;
                }
            });
        }
    };

    ta.prototype.reserveDelByNodeID = function(id, rf, sf) {
        var tf = this.createIdIndexMap();
        var uf;
        if (uf = tf[id]) {
            if (sf && (this.length - rf - 1) > uf || !sf && rf < uf) {
                var vf = {
                    value: uf,
                    order: String.fromCharCode(uf),
                    toString: function() {
                        return this.order
                    },
                    valueOf: function() {
                        return this.value
                    }
                };
                this.reserveDels.push(vf);
            }
        }
    };

    ta.prototype.reserveDelByNode = function(wf, xf, yf) {
        this.reserveDelByNodeID(Ze.get(wf), xf, yf);

    };
    ta.prototype.doDel = function() {
        if (!this.reserveDels.length) return;
        if (this.length < 0x10000) {
            var zf = this.reserveDels.sort(function(a, b) {
                return b - a
            });
        } else {
            var zf = this.reserveDels.sort(function(a, b) {
                return b - a
            });
        }
        for (var i = 0,
        l = zf.length; i < l; i++) {
            this.del(zf[i]);
        }
        this.reserveDels = [];
        this.idIndexMap = null;
    };
    ta.prototype.createIdIndexMap = function() {
        if (this.idIndexMap) {
            return this.idIndexMap;
        } else {
            var Af = this.idIndexMap = {};
            var Bf = this.nodes;
            for (var i = 0,
            l = Bf.length; i < l; i++) {
                var Cf = Bf[i];
                var id = Ze.get(Cf);
                Af[id] = i;
            }
            return Af;
        }
    };
    ta.prototype.del = function(Ff) {
        this.length--;
        if (this.only) {
            delete this.only;
        } else {
            var Gf = this.nodes.splice(Ff, 1)[0];
            if (this._first == Gf) {
                delete this._first;
                delete this._firstSourceIndex;
                delete this._firstSubIndex;
            }
            delete this.seen[Ze.get(Gf)];
        }
    };
    ta.prototype.delDescendant = function(Hf, If) {
        if (this.only) return;
        var Jf = Hf.nodeType;
        if (Jf != 1 && Jf != 9) return;
        if (Da.applewebkit2) return;
        if (!Hf.contains) {
            if (Jf == 1) {
                var Kf = Hf;
                Hf = {
                    contains: function(Lf) {
                        return Lf.compareDocumentPosition(Kf) & 8;
                    }
                };
            } else {
                Hf = {
                    contains: function() {
                        return true;
                    }
                };
            }
        }
        var Mf = this.nodes;
        for (var i = If + 1; i < Mf.length; i++) {
            if (Hf.contains(Mf[i])) {
                this.del(i);
                i--;
            }
        }
    };
    ta.prototype._add = function(Nf, Of) {
        var Qf = this.seen;
        var id = Ze.get(Nf);
        if (Qf[id]) return true;
        Qf[id] = true;
        this.length++;
        if (Of) this.nodes.unshift(Nf);
        else this.nodes.push(Nf);
    };
    ta.prototype.unshift = function(Rf) {
        if (!this.length) {
            this.length++;
            this.only = Rf;
            return
        }
        if (this.only) {
            var Sf = this.only;
            delete this.only;
            this.unshift(Sf);
            this.length--;
        }
        return this._add(Rf, true);
    };
    ta.prototype.push = function(Tf) {
        if (!this.length) {
            this.length++;
            this.only = Tf;
            return;
        }
        if (this.only) {
            var Uf = this.only;
            delete this.only;
            this.push(Uf);
            this.length--;
        }
        return this._add(Tf);
    };
    ta.prototype.first = function() {
        if (this.only) return this.only;
        if (this.nodes.length > 1) this.sort();
        return this.nodes[0];
    };
    ta.prototype.list = function() {
        if (this.only) return [this.only];
        this.sort();
        return this.nodes;
    };
    ta.prototype.string = function() {
        var Wf = this.only || this.first();
        return Wf ? wa.to('string', Wf) : '';
    };
    ta.prototype.bool = function() {
        return !! (this.length || this.only);
    };
    ta.prototype.number = function() {
        return + this.string();
    };
    ta.prototype.iterator = function(Xf) {
        this.sort();
        var Yf = this;
        if (!Xf) {
            var Zf = 0;
            return function() {
                if (Yf.only && Zf++==0) return Yf.only;
                return Yf.nodes[Zf++];
            };
        } else {
            var Zf = 0;
            return function() {
                var ag = Yf.length - (Zf++) - 1;
                if (Yf.only && ag == 0) return Yf.only;
                return Yf.nodes[ag];
            };
        }
    };
    var cg = function(dg) {
        dg = dg || this;
        var eg = dg.document;
        var ca = dg.undefined;
        dg.XPathExpression = function(fg) {
            if (!fg.length) {
                throw dg.Error('no expression');
            }
            var gg = this.lexer = Ea(fg);
            if (gg.empty()) {
                throw dg.Error('no expression');
            }
            this.expr = oa.parse(gg);
            if (!gg.empty()) {
                throw dg.Error('bad token: ' + gg.next());
            }
        };
        dg.XPathExpression.prototype.evaluate = function(hg, ig) {
            return new dg.XPathResult(this.expr.evaluate(new Ia(hg)), ig);
        };
        dg.XPathResult = function(jg, kg) {
            if (kg == 0) {
                switch (typeof jg) {
                case 'object':
                    kg++;
                case 'boolean':
                    kg++;
                case 'string':
                    kg++;
                case 'number':
                    kg++;
                }
            }
            this.resultType = kg;
            switch (kg) {
            case 1:
                this.numberValue = jg.isNodeSet ? jg.number() : +jg;
                return;
            case 2:
                this.stringValue = jg.isNodeSet ? jg.string() : '' + jg;
                return;
            case 3:
                this.booleanValue = jg.isNodeSet ? jg.bool() : !!jg;
                return;
            case 4:
            case 5:
            case 6:
            case 7:
                this.nodes = jg.list();
                this.snapshotLength = jg.length;
                this.index = 0;
                this.invalidIteratorState = false;
                break;
            case 8:
            case 9:
                this.singleNodeValue = jg.first();
                return;
            }
        };
        dg.XPathResult.prototype.iterateNext = function() {
            return this.nodes[this.index++]
        };
        dg.XPathResult.prototype.snapshotItem = function(i) {
            return this.nodes[i]
        };
        dg.XPathResult.ANY_TYPE = 0;
        dg.XPathResult.NUMBER_TYPE = 1;
        dg.XPathResult.STRING_TYPE = 2;
        dg.XPathResult.BOOLEAN_TYPE = 3;
        dg.XPathResult.UNORDERED_NODE_ITERATOR_TYPE = 4;
        dg.XPathResult.ORDERED_NODE_ITERATOR_TYPE = 5;
        dg.XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE = 6;
        dg.XPathResult.ORDERED_NODE_SNAPSHOT_TYPE = 7;
        dg.XPathResult.ANY_UNORDERED_NODE_TYPE = 8;
        dg.XPathResult.FIRST_ORDERED_NODE_TYPE = 9;
        eg.createExpression = function(lg) {
            return new dg.XPathExpression(lg, null);
        };
        eg.evaluate = function(mg, ng, _, og) {
            return eg.createExpression(mg, null).evaluate(ng, og);
        };
    };
    var pg;
    if (ea.targetFrame) {
        var qg = document.getElementById(ea.targetFrame);
        if (qg) pg = qg.contentWindow;
    }
    if (ea.exportInstaller) {
        window.install = cg;
    }
    if (!ea.hasNative || !ea.useNative) {
        cg(pg || window);
    }
})();
window.install();
'''



try:
    from qt4w.webdriver.webkitwebdriver import WebkitWebDriver
    from qt4w.util import JavaScriptError
    from qt4a.androiddriver.util import AndroidSpyError
    class AndroidWebDriver(WebkitWebDriver):
        '''Android中的WebDriver
        '''
        
        def __init__(self, webview):
            super(AndroidWebDriver, self).__init__(webview)
            self._usb_xpath_lib = False
            
        def eval_script(self, frame_xpaths, script):
            '''在指定frame中执行JavaScript，并返回执行结果（该实现需要处理js基础库未注入情况的处理）
            
            :param frame_xpaths: frame元素的XPATH路径，如果是顶层页面，怎传入“[]”
            :type frame_xpaths:  list
            :param script:       要执行的JavaScript语句
            :type script:        string
            '''
            try:
                return super(AndroidWebDriver, self).eval_script(frame_xpaths, script)
            except JavaScriptError as e:
                err_msg = e.message
                err_msg = err_msg.split('\n')[0]
                if 'INVALID_EXPRESSION_ERR: DOM XPath Exception 51' in err_msg:
                    # http://stackoverflow.com/questions/21472896/android-4-3-default-browser-xmldoc-evaluate-throws-invalid-expression-err-dom
                    # 可能是三星手机的bug，需要使用其它的XPath库
                    self.eval_script(frame_xpaths, xpath_lib_str)
                    self._usb_xpath_lib = True
                    return self.eval_script(frame_xpaths, script)
                elif 'Result Error:' in err_msg:
                    # 执行js超时，重试
                    time.sleep(5)
                    return self.eval_script(frame_xpaths, script)
                else:
                    raise e
            except AndroidSpyError as e:
                if 'java.lang.RuntimeException: Result Error:' in e.message:
                    # 一般是页面未加载完成，等待重试
                    time.sleep(2)
                    return self.eval_script(frame_xpaths, script)
                
except ImportError:
    pass


if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''Android测试桩代理
'''
from __future__ import print_function

import base64
import json
import re
import os
import six
import sys
import tempfile
import threading
import time

from qt4a.androiddriver.clientsocket import DirectAndroidSpyClient
from qt4a.androiddriver.devicedriver import DeviceDriver
from qt4a.androiddriver.util import logger, general_encode, SocketError, AndroidSpyError, ControlAmbiguousError, ControlExpiredError, ProcessExitError, QT4ADriverNotInstalled, Mutex, extract_from_zipfile


class EnumCommand(object):
    '''所有支持的命令字
    '''
    CmdHello = "Hello"
    CmdEnableDebug = "EnableDebug"
    CmdGetControl = "GetControl"
    CmdGetParent = "GetParent"
    CmdGetChildren = "GetChildren"
    CmdGetControlText = "GetControlText"
    CmdSetControlText = "SetControlText"
    CmdGetTextRect = "GetTextBoundingRect"
    CmdGetControlRect = "GetControlRect"
    CmdGetControlType = "GetControlType"
    CmdGetControlVisibility = "GetControlVisibility"
    CmdGetControlChecked = "GetControlChecked"
    CmdGetControlProgress = "GetControlProgress"
    CmdGetControlGravity = "GetControlGravity"
    CmdCaptureControl = "CaptureControl"
    CmdSendKey = "SendKey"
    CmdClick = "Click"
    CmdDrag = "Drag"
    CmdEnableSoftInput = "EnableSoftInput"
    CmdCloseActivity = "CloseActivity"
    CmdEvalScript = "EvalScript"
    CmdGetCurrentView = "GetCurrentView"
    CmdGetControlScrollRect = "GetControlScrollRect"
    CmdGetListViewInfo = "GetListViewInfo"
    CmdGetControlBackground = "GetControlBackground"
    CmdGetControlImageResource = "GetControlImageResource"
    CmdGetSelectedTabIndex = "GetSelectedTabIndex"
    CmdGetStaticFieldValue = "GetStaticFieldValue"
    CmdGetObjectFieldValue = "GetObjectFieldValue"
    CmdCallStaticMethod = "CallStaticMethod"
    CmdCallObjectMethod = "CallObjectMethod"
    CmdCallExternalMethod = "CallExternalMethod"  # 调用外部jar包中的方法
    CmdSetActivityPopup = "SetActivityPopup"
    CmdSetThreadPriority = "SetThreadPriority"
    CmdSetWebViewDebuggingEnabled = "SetWebViewDebuggingEnabled"


def install_qt4a_helper(adb, root_path):
    from qt4a.androiddriver.util import AndroidPackage, version_cmp
    qt4a_helper_package = 'com.test.androidspy'
    apk_path = os.path.join(root_path, 'QT4AHelper.apk')
    if adb.get_package_path(qt4a_helper_package):
        # 判断版本
        installed_version = adb.get_package_version(qt4a_helper_package)
        package = AndroidPackage(apk_path)
        install_version = package.version
        if version_cmp(install_version, installed_version) > 0:
            adb.install_apk(apk_path, True)
    else:
        adb.install_apk(apk_path)


def copy_android_driver(device_id_or_adb, force=False, root_path=None, enable_acc=True):
    '''测试前的测试桩拷贝
    '''
    from qt4a.androiddriver.adb import ADB

    if isinstance(device_id_or_adb, ADB):
        adb = device_id_or_adb
    else:
        adb = ADB.open_device(device_id_or_adb)

    if not root_path:
        current_path = os.path.abspath(__file__)
        if not os.path.exists(current_path) and '.egg' in current_path:
            # in egg
            egg_path = current_path
            while not os.path.exists(egg_path):
                egg_path = os.path.dirname(egg_path)
            assert(egg_path.endswith('.egg'))
            root_path = os.path.join(tempfile.mkdtemp(), 'tools')
            extract_from_zipfile(
                egg_path, 'qt4a/androiddriver/tools', root_path)
        else:
            root_path = os.path.join(os.path.dirname(
                os.path.abspath(__file__)), 'tools')
    dst_path = '/data/local/tmp/qt4a/'

    current_version_file = os.path.join(root_path, 'version.txt')
    f = open(current_version_file, 'r')
    current_version = int(f.read())
    f.close()

    if not force:
        version_file = dst_path + 'version.txt'
        version = adb.run_shell_cmd('cat %s' % version_file)

        if version and not 'No such file or directory' in version and current_version <= int(version):
            install_qt4a_helper(adb, root_path) # 避免QT4A助手被意外删除的情况
            # 不需要拷贝测试桩
            logger.warn('忽略本次测试桩拷贝：当前版本为%s，设备中版本为%s' %
                        (current_version, int(version)))
            return

    try:
        adb.chmod(dst_path[:-1], '777')
    except:
        pass

    rooted = adb.is_rooted()

    cpu_abi = adb.get_cpu_abi()
    print('Current CPU arch: %s' % cpu_abi)
    use_pie = False
    if adb.get_sdk_version() >= 21 and cpu_abi != 'arm64-v8a':
        use_pie = True

    file_list = [os.path.join(cpu_abi, 'droid_inject'), os.path.join(cpu_abi, 'libdexloader.so'), os.path.join(
        cpu_abi, 'screenkit'), os.path.join(cpu_abi, 'libandroidhook.so'), 'inject', 'AndroidSpy.jar', 'SpyHelper.jar', 'SpyHelper.sh']

    if cpu_abi == 'arm64-v8a':
        file_list.append(os.path.join(cpu_abi, 'droid_inject64'))
        file_list.append(os.path.join(cpu_abi, 'libdexloader64.so'))
        file_list.append('inject64')
    if adb.get_sdk_version() >= 21:
        file_list.append(os.path.join(cpu_abi, 'libandroidhook_art.so'))

    if rooted and adb.is_selinux_opened():
        # 此时如果还是开启状态说明关闭selinux没有生效,主要是三星手机上面
        adb.run_shell_cmd('rm -r %s' % dst_path, True)
        # adb.run_shell_cmd('chcon u:object_r:shell_data_file:s0 %slibdexloader.so' % dst_path, True)  # 恢复文件context，否则拷贝失败
        # adb.run_shell_cmd('chcon u:object_r:shell_data_file:s0 %slibandroidhook.so' % dst_path, True)

    for file in file_list:
        file_path = os.path.join(root_path, file)
        if use_pie and not '.' in file and os.path.exists(file_path + '_pie'):
            file_path += '_pie'
        if not os.path.exists(file_path):
            continue
        save_name = os.path.split(file)[-1]
        if save_name.endswith('_art.so'):
            save_name = save_name.replace('_art', '')
        adb.push_file(file_path, dst_path + save_name)

    adb.chmod('%sdroid_inject' % dst_path, 755)
    adb.chmod('%sinject' % dst_path, 755)
    adb.chmod('%sscreenkit' % dst_path, 755)
    adb.run_shell_cmd('ln -s %sscreenkit %sscreenshot' % (dst_path, dst_path))

    if cpu_abi == 'arm64-v8a':
        adb.chmod('%sdroid_inject64' % dst_path, 755)
        adb.chmod('%sinject64' % dst_path, 755)

    try:
        print(adb.run_shell_cmd('rm -R %scache' %
                                dst_path, rooted))  # 删除目录 rm -rf
    except RuntimeError as e:
        logger.warn('%s' % e)
    # logger.info(adb.run_shell_cmd('mkdir %scache' % (dst_path), True)) #必须使用root权限，不然生成odex文件会失败

    adb.mkdir('%scache' % (dst_path), 777)

    install_qt4a_helper(adb, root_path)

    version_file_path = os.path.join(root_path, 'version.txt')
    dst_version_file_path = dst_path + os.path.split(version_file_path)[-1]
    adb.push_file(version_file_path, dst_version_file_path + '.tmp')  # 拷贝版本文件

    if rooted and adb.is_selinux_opened():
        # 此时如果还是开启状态说明关闭selinux没有生效,主要是三星手机上面
        # 获取sdcars context
        if adb.get_sdk_version() >= 23:
            sdcard_path = adb.get_sdcard_path()
            result = adb.run_shell_cmd('ls -Z %s' % sdcard_path)
            # u:object_r:media_rw_data_file:s0 u:object_r:rootfs:s0
            pattern = re.compile(r'\s+(u:object_r:.+:s0)\s+')
            ret = pattern.search(result)
            if not ret:
                raise RuntimeError('get sdcard context failed: %s' % result)
            context = ret.group(1)
            logger.info('sdcard context is %s' % context)
            adb.run_shell_cmd('chcon %s %s' %
                              (context, dst_path), True)  # make app access
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %sSpyHelper.jar' % dst_path, True)
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %sSpyHelper.sh' % dst_path, True)
            # 不修改文件context无法加载so
            adb.run_shell_cmd(
                'chcon u:object_r:system_file:s0 %slibdexloader.so' % dst_path, True)
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %slibandroidhook.so' % dst_path, True)
            adb.run_shell_cmd('chcon %s %sAndroidSpy.jar' %
                              (context, dst_path), True)
            adb.run_shell_cmd('chcon %s %scache' % (context, dst_path), True)
        else:
            # 不修改文件context无法加载so
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %slibdexloader.so' % dst_path, True)
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %slibandroidhook.so' % dst_path, True)
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %scache' % dst_path, True)

    if rooted:
        if adb.get_sdk_version() < 24:
            # 7.0以上发现生成的dex与运行时生成的dex有差别，可能导致crash
            logger.info(adb.run_shell_cmd('sh %sSpyHelper.sh loadDex %sAndroidSpy.jar %scache' % (
                dst_path, dst_path, dst_path), rooted))
            adb.chmod('%scache/AndroidSpy.dex' % dst_path, 666)
    else:
        if not 'usage:' in adb.run_shell_cmd('sh %sSpyHelper.sh' % dst_path):
            adb.mkdir('%scache/dalvik-cache' % dst_path, 777)

    if rooted and adb.is_selinux_opened() and adb.get_sdk_version() >= 23:
        # 提升权限
        try:
            adb.list_dir('/system/bin/app_process32')
        except RuntimeError:
            adb.copy_file('/system/bin/app_process',
                          '%sapp_process' % dst_path)
        else:
            adb.copy_file('/system/bin/app_process32',
                          '%sapp_process' % dst_path)
        adb.chmod('%sapp_process' % dst_path, 755)
        adb.run_shell_cmd(
            'chcon u:object_r:system_file:s0 %sapp_process' % dst_path, True)

    adb.run_shell_cmd('mv %s %s' % (dst_version_file_path +
                                    '.tmp', dst_version_file_path), rooted)

    # 同步手机时间
    device_driver = DeviceDriver(adb)
    try:
        input_method = 'com.test.androidspy/.service.QT4AKeyboardService'
        device_driver.modify_system_setting(
            'secure', 'enabled_input_methods', input_method)
        device_driver.modify_system_setting(
            'secure', 'default_input_method', input_method)
        if enable_acc:
            device_driver.modify_system_setting('secure', 'enabled_accessibility_services',
                                                'com.test.androidspy/com.test.androidspy.service.QT4AAccessibilityService')
            device_driver.modify_system_setting(
                'secure', 'accessibility_enabled', 1)
    except:
        logger.exception('set default input method failed')
    try:
        device_driver.modify_system_setting('system', 'time_12_24', 24)
        device_driver.modify_system_setting(
            'system', 'screen_off_timeout', 600 * 1000)
    except:
        logger.exception('set system time failed')


class AndroidDriver(object):
    '''
    '''

    qt4a_path = '/data/local/tmp/qt4a'

    def __init__(self, device_driver, process_name, addr='127.0.0.1'):
        self._device_driver = device_driver
        self._adb = device_driver.adb
        self._process_name = process_name
        self._addr = addr
        self._process = {}  # 保存原始PID用于判断是否是新进程
        self._sync_lock = {}  # 同步锁
        self._max_block_time = 30  # 最长锁时间,防止卡死
        self._lock = threading.Lock()  # 创建AndroidDriver实例的互斥锁
        self._is_init = False
        self._client = None

    @staticmethod
    def create(process_name, device_or_driver):
        '''创建Driver实例

        :param process_name: 目标进程名
        :type  process_name: string
        :param device:       设备实例
        :type  device:       Device or DeviceDriver
        '''
        if hasattr(device_or_driver, '_device_driver'):
            device_driver = device_or_driver._device_driver
        else:
            device_driver = device_or_driver
        host_addr = '127.0.0.1'
        if not device_driver._is_local_device:
            host_addr = device_driver.adb.host_name
        return AndroidDriver(device_driver, process_name, addr=host_addr)

    @staticmethod
    def get_process_name_hash(process_name, device_id):
        '''根据进程名和设备id计算端口值
                结果范围：[5100, 5200)
        TODO: hash冲突的解决
        '''
        result = 0
        start_port = 15000
        port_range = 1000
        text = device_id + process_name
        for i in range(len(text)):
            c = text[i]
            asc = ord(c)
            result = 31 * result + asc

        if result > port_range:
            result %= port_range
        return result + start_port

    def _create_client(self):
        '''创建新的Client实例
        '''
        sock = self._adb.create_tunnel(self._process_name, 'localabstract')
        if sock == None:
            return None
        return DirectAndroidSpyClient(sock, timeout=360)

    def _safe_init_driver(self):
        '''多线程安全的初始化测试桩
        '''
        if not self._is_init:
            with Mutex(self._lock):
                if not self._is_init:
                    self._init_driver()
                    self._is_init = True

    def _init_driver(self):
        '''初始化测试桩
        '''
        self._client = self._create_client()
        if self._client != None:
            # 字段赋值
            self._process['name'] = self._process_name
            self._process['id'] = 0  # process id may change
            if self.hello() != None:
                self._process['id'] = self._adb.get_pid(self._process_name)
                return

        timeout = 20
        time0 = time.time()
        proc_exist = False
        while time.time() - time0 < timeout:
            if not proc_exist:
                pid = self._adb.get_pid(self._process_name)
                if pid > 0:
                    proc_exist = True
                    self._process['name'] = self._process_name
                    self._process['id'] = pid
                    break

            time.sleep(1)

        if not proc_exist:
            raise RuntimeError('进程：%s 在%d秒内没有出现' %
                               (self._process_name, timeout))

        inject_file = 'inject'
        if self._adb.is_app_process64(pid if self._adb.is_rooted() else self._process_name):
            # 64 bit process
            inject_file += '64'
        timeout = 30

        try:
            if self._adb.is_art():
                # Android 5.0上发现注入容易导致进程退出
                self._wait_for_cpu_low(20, 10)

            time0 = time.time()
            cmdline = '%s/%s %s' % (self._get_driver_root_path(),
                                    inject_file, self._process_name)
            while time.time() - time0 < timeout:
                if self._adb.is_rooted():
                    ret = self._adb.run_shell_cmd(
                        cmdline, True, timeout=120, retry_count=1)
                else:
                    ret = self._adb.run_as(
                        self._process_name, cmdline, timeout=120, retry_count=1)
                logger.debug('inject result: %s' % ret)
                if 'not found' in ret:
                    raise QT4ADriverNotInstalled(
                        'QT4A driver damaged, please reinstall QT4A driver')
                if 'Inject Success' in ret:
                    break
                elif 'Operation not permitted' in ret:
                    # 可能是进程处于Trace状态
                    pid = self._adb.get_pid(self._process_name)
                    status = self._adb.get_process_status(pid)
                    tracer_pid = int(status['TracerPid'])
                    if tracer_pid > 0:
                        if int(status['PPid']) == tracer_pid:
                            # 使用TRACEME方式防注入
                            raise Exception('应用使用了防注入逻辑,注入失败')
                        logger.warn('TracerPid is %d' % tracer_pid)
                        self._adb.kill_process(tracer_pid)
                time.sleep(1)

        except RuntimeError as e:
            logger.exception('%s\n%s' % (e, self._adb.run_shell_cmd('ps')))
            if self._adb.is_rooted():
                logger.info(self._adb.dump_stack(self._process_name))
            raise e
        timeout = 10
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self._client == None:
                self._client = self._create_client()
            if self._client != None and self.hello() != None:
                return
            time.sleep(0.1)
        raise RuntimeError('连接测试桩超时')

    def _get_driver_root_path(self):
        '''获取驱动文件根目录
        '''
        if self._adb.is_rooted():
            return self.__class__.qt4a_path
        package_name = self._process_name
        if ':' in package_name:
            package_name = package_name.split(':')[0]
        if not self._device_driver.is_debug_package(package_name):
            raise NotImplementedError(
                'Non root device doesn\'t support release package')
        root_path = '/data/data/%s/qt4a' % package_name  # 使用应用目录作为根目录
        result = self._adb.run_as(package_name, 'ls -l %s/inject' % root_path)
        if not 'No such file or directory' in result:
            return root_path

        # 拷贝驱动文件
        self._adb.run_as(package_name, 'mkdir %s' % root_path)
        self._adb.run_as(package_name, 'mkdir %s/cache' % root_path)
        file_list = ['AndroidSpy.jar']  # , 'cache/AndroidSpy.dex'
        if self._adb.is_app_process64(package_name):
            file_list.extend(
                ['droid_inject64', 'inject64', 'libdexloader64.so'])
        else:
            file_list.extend(
                ['droid_inject', 'inject', 'libdexloader.so', 'libandroidhook.so'])
        for filename in file_list:
            self._adb.run_as(package_name, 'cp %s/%s %s/%s' %
                             (self.__class__.qt4a_path, filename, root_path, filename))
        # self._device.adb.run_as(package_name, 'sh %s/SpyHelper.sh loadDex %s/AndroidSpy.jar %s/cache' % (self.__class__.qt4a_path, self.__class__.qt4a_path, root_path))
        return root_path

    def _wait_for_cpu_low(self, max_p_cpu, max_t_cpu, timeout=20, interval=0.5):
        '''等待待注入进程CPU使用率降低到max_p_cpu，线程CPU使用率降低到max_t_cpu

        :param max_p_cpu: 进程占用的CPU
        :type max_p_cpu:  int
        :param max_t_cpu: 线程占用的CPU
        :type max_t_cpu:  int
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            ret = self._adb.get_process_cpu(self._process_name)
            if ret != None:
                p_cpu, t_cpu = ret
                logger.debug('current cpu: %d, %d' % (p_cpu, t_cpu))
                if p_cpu < max_p_cpu and t_cpu < max_t_cpu:
                    return
            time.sleep(interval)

    def close(self):
        '''关闭连接
        '''
        if self._client:
            self._client.close()
            self._client = None

    def _get_lock(self, thread_id):
        '''获取锁
        '''
        if not thread_id in self._sync_lock:
            self._sync_lock[thread_id] = threading.Event()
            self._sync_lock[thread_id].set()
        return self._sync_lock[thread_id]

    def suspend_thread(self, thread_id, max_wait_time=30):
        '''阻塞
        '''
        self._max_block_time = max_wait_time
        self._get_lock(thread_id).clear()  # Reset the internal flag to false

    def resume_thread(self, thread_id):
        '''解锁
        '''
        self._get_lock(thread_id).set()  # Set the internal flag to true

    def _wait_for_event(self, thread_id, timeout):
        '''等待事件
        '''
        if self._get_lock(thread_id).is_set():
            return
        time0 = time.time()
        ret = self._get_lock(thread_id).wait(timeout)
        if ret:
            cost_time = time.time() - time0
            logger.debug('thread %s wait %s S' % (thread_id, cost_time))
        else:
            # 超时
            logger.warn('thread %s wait for sync event timeout' % thread_id)
            self.resume_thread(thread_id)  # 防止出错后一直等待
        return ret

    def send_command(self, cmd_type, **kwds):
        '''发送命令
        '''
        curr_thread_id = threading.current_thread().ident
        self._wait_for_event(curr_thread_id, self._max_block_time)

        if cmd_type != EnumCommand.CmdHello:
            self._safe_init_driver()  # 确保测试桩连接正常

        result = self._client.send_command(cmd_type, **kwds)
        if result == None:
            pid = self._adb.get_pid(self._process['name'])
            if pid > 0 and pid != self._process['id']:
                # 新进程，需要重新注入
                logger.info('process %s id changed: %d %d' %
                            (self._process['name'], self._process['id'], pid))
                self._is_init = False  # 需要重新初始化
                self._process['id'] = pid
                return self.send_command(cmd_type, **kwds)
            elif pid == 0:
                raise ProcessExitError('被测进程已退出，确认是否发生Crash')
            elif cmd_type != EnumCommand.CmdHello:
                # hello包不重试
                logger.debug('socket error, try to reconnect')
                for _ in range(3):
                    # 为防止由于设备短暂失联导致的连接断开，这里调一次adb forward
                    self._client = self._create_client()
                    result = self._client.send_command(cmd_type, **kwds)
                    if result != None:
                        return result
                raise SocketError('Connect Failed')
            else:
                raise SocketError('Connect Failed')
        if 'Error' in result:
            if result['Error'] == u'控件已失效':
                raise ControlExpiredError(result['Error'])
            elif result['Error'] == u'控件类型错误':
                control_type = self.get_control_type(kwds['Control'])
                raise TypeError('%s,当前控件类型为:%s' %
                                (result['Error'], control_type))
            else:
                raise AndroidSpyError(result['Error'])
        return result

    def hello(self):
        '''确认Server身份
        '''
        for _ in range(3):
            try:
                result = self.send_command(EnumCommand.CmdHello)
            except SocketError:
                time.sleep(1)
                continue
            except AndroidSpyError:
                logger.exception('init error')
                time.sleep(2)
                continue
            # self._enable_debug()
            if not 'Result' in result:
                raise RuntimeError('Server error')
            items = result['Result'].split(':')
            if len(items) > 0 and items[-1].isdigit():
                if self._process['id'] and int(items[-1]) != self._process['id']:
                    raise RuntimeError(
                        'Server pid not match %s' % result['Result'])
            return result['Result']

    def _enable_debug(self):
        '''启用Debug模式
        '''
        self.send_command(EnumCommand.CmdEnableDebug)

    def get_control(self, activity, parent, locator, get_position=False):
        '''获取控件hashcode
        '''
        kwds = {}
        if activity:
            kwds['Activity'] = activity
        if parent:
            kwds['Parent'] = parent
        kwds['Locator'] = locator
        if get_position:
            kwds['GetPosition'] = get_position
        try:
            result = self.send_command(EnumCommand.CmdGetControl, **kwds)
        except ControlExpiredError as e:
            raise e
        except AndroidSpyError as e:
            err_msg = e.args[0].strip()
            err_msg = general_encode(err_msg)
            if '找到重复控件' in err_msg or 'Multiple controls found' in err_msg:
                qpath = '/'.join(' && '.join([(key + it[key][0] + '"' + str(
                    it[key][1]) + '"') for key in it.keys()]) for it in locator)
                qpath = general_encode(qpath)
                err_msg = '定位到重复控件：/%s\n%s' % (qpath, '\n'.join(err_msg.split('\n')[1:]))
                raise ControlAmbiguousError(err_msg)
            elif '查找控件失败' in err_msg or 'Control not found' in err_msg:
                if get_position and ':' in err_msg:
                    _, pos = err_msg.split(':')
                    return int(pos)
            return 0
        if 'Result' in result:
            return result['Result']
        return 0

    def get_parent(self, control):
        '''获取父控件
        '''
        result = self.send_command(EnumCommand.CmdGetParent, Control=control)
        return result['Result']

    def get_children(self, control):
        '''获取子节点列表
        '''
        result = self.send_command(EnumCommand.CmdGetChildren, Control=control)
        return result['Result']

    def get_control_text(self, control, html_format=False):
        '''获取控件文本
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlText, Control=control, UseHtml=html_format)
        result = result['Result']
        if html_format:
            # 处理html编码
            from six.moves.html_parser import HTMLParser
            result = HTMLParser().unescape(result)
        return result

    def set_control_text(self, control, text):
        '''设置控件文本
        '''
        result = self.send_command(
            EnumCommand.CmdSetControlText, Control=control, Text=text)
        return result['Result']

    def get_text_rect(self, control, start_offset, end_offset):
        '''获取文本框中指定字符串的区域坐标
        '''
        result = self.send_command(
            EnumCommand.CmdGetTextRect, Control=control, StartOffset=start_offset, EndOffset=end_offset)
        return result['Result']

    def get_control_type(self, control, include_baseclass=False):
        '''获取控件类型
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlType, Control=control, IncludeBaseClass=include_baseclass)
        if not include_baseclass:
            return result['Result']
        else:
            return result['Result'].split(',')

    def get_control_rect(self, control):
        '''获取控件位置信息
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlRect, Control=control)
        return result['Result']

    def get_control_visibility(self, control):
        '''获取控件可见性
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlVisibility, Control=control)
        return result['Result']

    def get_control_checked(self, control):
        '''获取CommpandButton控件是否选择
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlChecked, Control=control)
        return result['Result']

    def get_control_gravity(self, control):
        '''获取控件位置信息
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlGravity, Control=control)
        return result['Result']

    def get_control_progress(self, control):
        '''获取ProgressBar控件的进度
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlProgress, Control=control)
        return result['Result']

    def get_control_background(self, control):
        '''获取控件背景资源id
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlBackground, Control=control)
        return result['Result']

    def get_control_scroll_rect(self, control):
        '''获取控件滚动区域范围
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlScrollRect, Control=control)
        return result['Result']

    def get_control_image_resource(self, control):
        '''获取控件Image资源信息
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlImageResource, Control=control)
        return result['Result']

    def get_listview_info(self, control, timeout=20):
        '''获取AbsListView类型的相关信息
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            # 过滤掉一些明显错误的数据
            result = self.send_command(
                EnumCommand.CmdGetListViewInfo, Control=control)
            result = result['Result']
            if (result['Count'] > 0 and result['FirstPosition'] >= result['Count']) or \
                    result['LastPosition'] > result['Count']:  # 最后一项>=总项数 {"LastPosition":0,"FirstPosition":0,"Count":0}
                time.sleep(0.03)
                continue
            return result
        raise RuntimeError('获取ListView控件信息错误')

    def get_selected_tab_index(self, control):
        '''获取TabWidget控件当前所选项
        '''
        result = self.send_command(
            EnumCommand.CmdGetSelectedTabIndex, Control=control)
        return result['Result']

    def capture_control(self, control):
        result = self.send_command(
            EnumCommand.CmdCaptureControl, Control=control)
        data = result['Result']
        if not isinstance(data, bytes):
            data = data.encode()
        data = base64.decodestring(data)
        return data

    def send_key(self, key_list):
        '''发送按键，只允许单个按键或组合键
        '''
        from qt4a.androiddriver.util import KeyCode
        if isinstance(key_list, str):
            return self.send_keys(key_list)
        elif isinstance(key_list, int):
            # 将整型按键转为list
            key_list = [key_list]
        if len(key_list) == 0 or len(key_list) > 3:
            raise RuntimeError('按键错误')
        keys = ''
        for key in key_list:
            keys += str(key) + ','
        keys = keys[:-1]
        if keys == str(KeyCode.KEYCODE_HOME) or keys == str(KeyCode.KEYCODE_BACK):  # Home键需要特殊处理
            return self._device_driver.send_key(int(keys))
        else:
            return self.send_command(EnumCommand.CmdSendKey, Keys=keys)

    def send_keys(self, text):
        '''发送字符串，目前不支持中文
        '''
        from qt4a.androiddriver.util import KeyCode
        key_list = KeyCode.get_key_list(text)
        for key in key_list:
            self.send_key(key)

    def click(self, control, x, y, sleep_time=0):
        if x < 0 or y < 0:
            raise RuntimeError('坐标错误：(%d, %d)' % (x, y))

        try:
            if not control:
                result = self.send_command(
                    EnumCommand.CmdClick, X=x, Y=y, SleepTime=int(sleep_time * 1000))
            else:
                result = self.send_command(
                    EnumCommand.CmdClick, Control=control, X=x, Y=y, SleepTime=int(sleep_time * 1000))
            # print (result['Result'])
            return result['Result']
        except AndroidSpyError as e:
            if str(e).find('java.lang.SecurityException') >= 0:
                # 有时操作过快会出现该问题
                return False
            elif 'ErrControlNotShown' in str(e):
                return False
            elif str(e) == '控件已失效':
                # 认为点击成功了
                pic_name = '%s.png' % int(time.time())
                logger.warn('%s %s' % (e, pic_name))
                self._device_driver.take_screen_shot(pic_name)
                return True
            else:
                raise e

    def drag(self, x1, y1, x2, y2, count=5, wait_time=40, send_down_event=True, send_up_event=True):
        '''滑动
        :param x1: 起始横坐标
        :type x1:  int
        :param y1: 起始纵坐标
        :type y1:  int
        :param x2: 终点横坐标
        :type x2:  int
        :param y2: 终点纵坐标
        :type y2:  int
        :param count: 步数
        :type count: int
        :param wait_time: 滑动过程中每步之间的间隔时间，ms
        :type wait_time: int
        :param send_down_event: 是否发送按下消息
        :type send_down_event: boolean
        :param send_up_event: 是否发送弹起消息
        :type send_up_event: boolean 
        '''
        if y1 != y2 and abs(y1 - y2) < 60:  # 此时滑动距离会很短
            # 保证最短滑动距离为40，在索尼ce0682上发现小于40时会变成点击 三星9300上发现有时60像素以下滑动不起来
            m = (y1 + y2) // 2
            if y1 - y2 > 0:
                d = 30
            else:
                d = -30
            y1 = m + d  # TODO:坐标合法性判断
            y2 = m - d
        for _ in range(3):
            try:
                result = self.send_command(EnumCommand.CmdDrag, X1=x1, Y1=y1, X2=x2, Y2=y2,
                                           StepCount=count,
                                           SleepTime=wait_time,
                                           SendDownEvent=send_down_event,
                                           SendUpEvent=send_up_event)
            except AndroidSpyError as e:
                if str(e).find('java.lang.SecurityException') >= 0:
                    logger.info('java.lang.SecurityException,current activity:%s' %
                                self._device_driver.get_current_activity())  # 检测是否有弹窗
                    # 有时操作过快会出现该问题
                    time.sleep(0.1)
                else:
                    raise e
            else:
                return True
        logger.error('drag (%s, %s, %s, %s) failed' % (x1, y1, x2, y2))
        return False

    def enable_soft_input(self, control, enable=False):
        '''启用/禁止软键盘
        '''
        self.send_command(EnumCommand.CmdEnableSoftInput,
                          Control=control, Enable=enable)

    def _close_activity(self, activity):
        '''关闭Activity
        '''
        result = self.send_command(
            EnumCommand.CmdCloseActivity, Activity=activity)
        return result['Result']

    def close_activity(self, activity):
        '''关闭Activity
        '''
        try:
            return self._close_activity(activity)
        except ProcessExitError:
            logger.warn('close_activity error: process %s not exist' %
                        self._process_name)
            return -3  # 进程退出

    def eval_script(self, control, frame_xpaths, script):
        '''执行JavaScript
        '''
        result = self.send_command(
            EnumCommand.CmdEvalScript, Control=control, Script=script, Frame=frame_xpaths)
        return result['Result']

    def get_static_field_value(self, class_name, field_name):
        '''获取类`class_name`中静态字段`field_name`的值
        '''
        result = self.send_command(
            EnumCommand.CmdGetStaticFieldValue, ClassName=class_name, FieldName=field_name)
        return result['Result']

    def get_object_field_value(self, control, field_name):
        '''获取对象中字段`field_name`的值
        '''
        result = self.send_command(
            EnumCommand.CmdGetObjectFieldValue, Control=control, FieldName=field_name)
        return result['Result']

    def call_static_method(self, class_name, method_name, control=None, ret_type='', *args):
        '''调用静态方法
        '''
        kwargs = {}
        if control:
            kwargs['Control'] = control
        result = self.send_command(EnumCommand.CmdCallStaticMethod, ClassName=class_name,
                                   Method=method_name, RetType=ret_type, Args=args, **kwargs)
        return result['Result']

    def call_object_method(self, control, inner_obj, method_name, ret_type='', *args):
        '''调用对象方法
        '''
        result = self.send_command(EnumCommand.CmdCallObjectMethod, Control=control,
                                   InnerObject=inner_obj, Method=method_name, RetType=ret_type, Args=args)
        return result['Result']

    def call_external_method(self, jar_path, cls_name, add_to_classloader=False, **kwds):
        '''调用外部jar包中的方法
        '''
        result = self.send_command(EnumCommand.CmdCallExternalMethod, JarPath=jar_path,
                                   ClsName=cls_name, AddToClassLoader=add_to_classloader, Args=kwds)
        return result['Result']

    def set_activity_popup(self, activity, popup=False):
        '''禁用某些Activity自动弹出
        '''
        result = self.send_command(
            EnumCommand.CmdSetActivityPopup, Activity=activity, Popup=popup)
        return result['Result']

    def _format_control_tree(self, result, indent=0):
        '''格式化控件树
        '''
        if not result:
            return ''
        id = result.pop('Id')
        padding = '+--' * indent + '+'
        ret = '%s%s (' % (padding, id)
        rect = result.pop('Rect')
        bounding_rect = '(%s, %s, %s, %s)' % (
            rect['Left'], rect['Top'], rect['Width'], rect['Height'])
        children = result.pop('Children')

        for key in result.keys():
            ret += '%s: %s, ' % (key, result[key])
        ret += 'BoundingRect: %s)\n' % bounding_rect
        for child in children:
            ret += self._format_control_tree(child, indent + 1)
        return ret

    def _get_control_tree(self, activity, index):
        '''获取控件树
        '''
        return self.send_command('GetControlTree', Activity=activity, Index=index)['Result']

    def get_control_tree(self, activity, index=-1):
        '''获取控件树
        '''
        result = self._get_control_tree(activity, index)
        if len(activity) > 0:
            return self._format_control_tree(result)
        else:
            output = ''
            for activity in result.keys():
                output += '%s:\n' % activity
                output += self._format_control_tree(result[activity])
            return output

    def set_thread_priority(self, priority):
        '''设置测试线程优先级

        :param priority: 要设置的优先级
        :type priority:  EnumThreadPriority
        '''
        self.send_command(EnumCommand.CmdSetThreadPriority, Priority=priority)

    def set_webview_debugging_enabled(self, control, enable=True):
        '''设置WebView调试状态

        :param control: WebView控件hashcode
        :type  control: int/long
        :param enable: 是否开启WebView调试
        :type  enable: bool
        '''
        self.send_command(
            EnumCommand.CmdSetWebViewDebuggingEnabled, Control=control, Enable=True)


if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''客户端Socket连接
'''

from __future__ import unicode_literals

import json
import six
import socket
import select
import time
import threading
from qt4a.androiddriver.util import logger, time_clock

def is_tcp_server_opened(addr, port):
    '''判断TCP服务是否连接正常
    '''
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.connect((addr, port))
        sock.close()
        return True
    except socket.error:
        return False


class TCPSocketClient(object):
    '''TCPSocket客户端
    '''
    def __init__(self, addr, port, timeout=60):
        self._sock = None
        self._addr = addr
        self._port = port
        self._connect = False
        self._timeout = timeout

    @staticmethod
    def server_opened(port, addr='127.0.0.1'):
        '''判断服务是否已经打开
        '''
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect((addr, port))
            sock.send(b'{"Cmd":"Hello","Seq":1}\n')  # 有时connect会成功，但是send才会发现失败
            sock.recv(4096)
            return True
        except socket.error as e:
            logger.debug('server_opened (%s, %s) %s' % (addr, port, e))
            if e.args[0] == 10035: return True
            return False

    def pre_connect(self, timeout=5):
        '''预连接，adb forward调用后不能保证连接一定可用
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            try:
                if self.hello() != None:
                    return True
            except socket.error as e:
                # logger.warn(e)
                time.sleep(0.5)  # 尝试频率没必要太频繁
        raise RuntimeError('Socket连接失败')

    def hello(self):
        return self.send('Hello')

    def connect(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # socket.setdefaulttimeout(120)
        for i in range(3):
            try:
                self._sock.connect((self._addr, self._port))
                self._sock.settimeout(self._timeout)
                self._connect = True
                return True
            except Exception as e:
                logger.warn('connect (%s, %s) error: %s' % (self._addr, self._port, e))
                if i == 2: return False  # 连接失败
                time.sleep(1)

    def close(self):
        if self._sock:
            self._sock.close()
            self._sock = None
    
    def recv(self, buff_size):
        return self._sock.recv(buff_size)
        
    def send(self, data):
        if not self._connect:
            if not self.connect(): return None
        try:
            self._sock.send(data.encode('utf8'))
        except socket.error as e:
            logger.info('发送%r错误： %s' % (data, e))
            self._sock.close()
            self._connect = False
            return None
        
        expect_len = self.recv(8)
        if not expect_len: return None

        expect_len = int(expect_len, 16) + 1
        recv_buff = ''
        max_utf8_length = 6

        while len(recv_buff) < expect_len:
            buff = self.recv(expect_len - len(recv_buff))

            if not buff:
                logger.warn('Socket closed when recv rsp for %r' % data)
            
            while True:
                try:
                    recv_buff += buff.decode('utf8')
                    break
                except UnicodeDecodeError:
                    if len(buff) > max_utf8_length:
                        for i in range(max_utf8_length):
                            this_buff = buff[:i - max_utf8_length]
                            try:
                                recv_buff += this_buff.decode('utf8')
                            except UnicodeDecodeError:
                                pass
                            else:
                                buff = buff[i - max_utf8_length:]
                                break
                        else:
                            raise RuntimeError('Invalid utf-8 bytes: %r' % buff)
                    buff += self.recv(1)

        return recv_buff

class AndroidSpyClient(TCPSocketClient):
    '''AndroidSpy客户端
    '''
    def __init__(self, port, addr='127.0.0.1', enable_log=True, timeout=20):
        super(AndroidSpyClient, self).__init__(addr, port, timeout)
        self._seq = 0
        self._enable_log = enable_log
        self._lock = threading.Lock()  # TODO: 多线程互斥

    @property
    def seq(self):
        self._seq += 1
        return self._seq

    def send_command(self, cmd_type, **kwds):
        '''send command
        '''
        packet = {}
        packet['Cmd'] = cmd_type
        packet['Seq'] = self.seq
        for key in kwds.keys():
            packet[key] = kwds[key]
        data = json.dumps(packet) + "\n"
        if six.PY2 and isinstance(data, unicode):
            data = data.encode('utf8')

        time0 = time_clock()
        self._lock.acquire()
        time1 = time_clock()
        delta = time1 - time0
        if self._enable_log and delta >= 0.05: logger.info('send wait %s S' % delta)
        if self._enable_log: logger.debug('send: %s' % (data[:512].strip()))

        time0 = time_clock()
        try:
            result = self.send(data)
        except Exception as e:
            # 避免因异常导致死锁
            logger.exception('send %r error: %s' % (data, e))
            result = None
        self._lock.release()  # 解锁
        if not result: return None

        time1 = time_clock()
        try:
            rsp = json.loads(result)
        except:
            logger.error('json error: %r' % (result))
            raise
        else:
            if self._enable_log: 
                delta = int(1000 * (time1 - time0))
                if 'HandleTime' in rsp: delta -= rsp['HandleTime']
                logger.debug('recv: [%d]%s\n' % (delta, result[:512].strip()))
            return rsp
        
    def hello(self):
        return self.send_command('Hello')

class DirectAndroidSpyClient(AndroidSpyClient):
    '''走直连方式的AndroidSpy客户端
    '''
    def __init__(self, sock, enable_log=True, timeout=20):
        super(DirectAndroidSpyClient, self).__init__(0, enable_log=enable_log, timeout=timeout)
        self._sock = sock
        self._connect = True
    
    def recv(self, buff_size):
        time0 = time.time()
        while time.time() - time0 < self._timeout:
            try:
                ret = self._sock.recv(buff_size)
                if ret != None: return ret
            except socket.timeout:
                pass

            time.sleep(0.001)
            continue

        raise socket.timeout('recv data timeout')
                

if __name__ == '__main__':
    pass




# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''andrcontrols模块单元测试
'''

try:
    from unittest import mock
except:
    import mock
import unittest

from qt4a.andrcontrols import Window, TextView, EditText
from qt4a.qpath import QPath

class MyWindow(Window):
    '''demo窗口
    '''
    Activity = 'com.tencent.demo.activity.MainActivity'
    Process = 'com.tencent.demo'
    
    def __init__(self, app, wait_activity=False):
        super(MyWindow, self).__init__(app, wait_activity=wait_activity)
        self.update_locator({'标题': {'type': TextView, 'root': self, 'locator': QPath('/Id="title"')},
                            '输入框': {'type': EditText, 'root': self, 'locator': QPath('/Id="edit"')},
                            })
        
class TestWindow(unittest.TestCase):
    '''Window类测试用例
    '''
    
    def _get_device(self):
        ADB.is_rooted = mock.Mock(return_value=False)
        ADB.run_shell_cmd = mock.Mock(side_effect=mock_run_shell_cmd)
        adb_backend = LocalADBBackend('127.0.0.1', 'test')
        adb = ADB(adb_backend)
        return Device(adb)
    
    def test_wait_for_exist(self):
        pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''重打包
'''

import six
import json
import logging
import os
import shutil
import sys
import subprocess
import tempfile
from qt4a.apktool.apkfile import APKFile
from qt4a.apktool.manifest import AndroidManifest


class MergeDexError(RuntimeError):
    '''合并dex失败错误
    '''
    pass


class TooManyMethodsError(MergeDexError):
    '''方法数超标
    '''
    pass


class OutOfMemoryError(RuntimeError):
    '''内存超标
    '''
    pass


def general_decode(s):
    if not isinstance(s, bytes):
        return s
    try:
        return s.decode('utf8')
    except:
        return s.decode('gbk')


def get_apk_signature(rsa_file_path):
    '''获取应用签名
    '''
    if not os.path.exists(rsa_file_path):
        raise RuntimeError('Rsa file %s not exist' % rsa_file_path)
    cur_path = os.path.dirname(os.path.abspath(__file__))
    jar_path = os.path.join(cur_path, 'tools', 'apkhelper.jar')
    return os.popen('java -jar %s getSignature %s' % (jar_path, rsa_file_path)).read()


def dex2jar(dex_path, jar_path):
    '''将dex转换为jar
    '''
    if not os.path.exists(dex_path):
        raise RuntimeError('dex %s not exist' % dex_path)
    cur_path = os.path.dirname(os.path.abspath(__file__))
    tools_path = os.path.join(cur_path, 'tools')
    class_paths = []
    for it in os.listdir(tools_path):
        if it.endswith('.jar'):
            class_paths.append(os.path.join(tools_path, it))
    # -Xms512m -Xmx1024m
    sep = ';' if sys.platform == 'win32' else ':'
    cmdline = ['java', '-Xmx1024m', '-cp', sep.join(class_paths), 'com.googlecode.dex2jar.tools.Dex2jarCmd',
               '-f', '-o', jar_path, dex_path]
    logging.info(' '.join(cmdline))
    proc = subprocess.Popen(
        cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    result = proc.communicate()[1]
    if not b'->' in result:
        raise RuntimeError('convert dex %s to jar failed: %s' %
                           (dex_path, result))


def jar2dex(jar_path, dex_path, max_heap_size=0):
    '''jar转换为dex
    '''
    if not os.path.exists(jar_path):
        raise RuntimeError('jar %s not exist' % jar_path)
    cur_path = os.path.dirname(os.path.abspath(__file__))
    dx_path = os.path.join(cur_path, 'tools', 'dx.jar')
    # cmdline = 'java -Xmx1024m -cp "%s" com.android.dx.command.Main --dex --force-jumbo --output="%s" "%s"' % (dx_path, dex_path, jar_path)
    memory = 1024
    if max_heap_size:
        memory = max_heap_size
    elif os.path.getsize(jar_path) >= 8 * 1024 * 1024:
        memory = 2048  # 此时必须用64位java
    cmdline = ['java', '-Xmx%dm' % memory, '-cp', dx_path, 'com.android.dx.command.Main',
               '--dex', '--force-jumbo', '--output=%s' % dex_path, jar_path]
    logging.info(' '.join(cmdline))

    proc = subprocess.Popen(
        cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    result = proc.communicate()[1]
    if result:
        err_msg = result
        if sys.platform == 'win32':
            err_msg = err_msg.decode('gbk')
        else:
            err_msg = err_msg.decode('utf8')
        if proc.returncode:
            if 'java.lang.OutOfMemoryError' in err_msg:
                raise OutOfMemoryError(err_msg)
            else:
                raise RuntimeError(
                    'Convert jar %s to dex failed: %s' % (jar_path, err_msg))
        else:
            logging.warning(err_msg)


def rebuild_dex_with_jumbo(dex_path, max_heap_size=0):
    '''使用jumbo模式重新编译dex
    '''
    jar_path = tempfile.mktemp('.jar')
    dex2jar(dex_path, jar_path)
    jar2dex(jar_path, dex_path, max_heap_size)
    os.remove(jar_path)


def merge_dex(dst_dex, src_dexs, max_heap_size=0):
    '''合并dex
    '''
    cur_path = os.path.dirname(os.path.abspath(__file__))
    jar_path = os.path.join(cur_path, 'tools', 'apkhelper.jar')
    cmdline = ['java', '-jar', jar_path, 'mergeDex', dst_dex]
    cmdline.extend(src_dexs)
    logging.info(' '.join(cmdline))
    proc = subprocess.Popen(
        cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    result = proc.communicate()
    if not b'Took' in result[0]:
        if b'non-jumbo instruction' in result[1]:
            # 应用未使用jumbo模式编译，这里做一次转换
            logging.warn('change dex to jumbo mode')
            rebuild_dex_with_jumbo(src_dexs[0], max_heap_size)
            return merge_dex(dst_dex, src_dexs, max_heap_size)
        elif b'DexIndexOverflowException' in result[1]:
            raise TooManyMethodsError('Merge dex failed: %s' % result[1])
        else:
            raise MergeDexError('Merge dex failed: %s' % result[1])


def resign_apk(apk_path):
    '''重签名应用
    '''
    if not os.path.exists(apk_path):
        raise RuntimeError('apk %s not exist' % apk_path)
    cur_path = os.path.dirname(os.path.abspath(__file__))
    key_file_path = os.path.join(cur_path, 'tools', 'qt4a.keystore')
    save_path = apk_path[:-4] + '-signed.apk'
    # 'jarsigner -digestalg SHA1 -sigalg MD5withRSA -keystore %s -signedjar %s %s qt4a' % (key_file_path, save_path, apk_path)
    cmdline = ['jarsigner', '-digestalg', 'SHA1', '-sigalg', 'MD5withRSA', '-keystore',
               key_file_path, '-signedjar', save_path, apk_path, 'qt4a']
    logging.info(' '.join(cmdline))
    proc = subprocess.Popen(cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate(b'test@123')
    if out:
        logging.info('jarsigner: ' + general_decode(out))
    if err:
        logging.warn('jarsigner: ' + general_decode(err))
    return save_path


def repack_apk(apk_path_or_list, provider_name, merge_dex_path, activity_list=None, res_file_list=None, debuggable=None, vm_safe_mode=None, max_heap_size=0, force_append=False):
    '''
    '''
    if not isinstance(apk_path_or_list, list):
        apk_path_or_list = [apk_path_or_list]
    elif len(apk_path_or_list) == 0:
        raise ValueError('apk path not specified')

    apk_file_list = []
    signature_dict = {}
    for it in apk_path_or_list:
        apk_file = APKFile(it)
        apk_file_list.append(apk_file)
        manifest = AndroidManifest(apk_file)
        if debuggable != None:
            manifest.debuggable = debuggable  # 修改debuggable属性
        if vm_safe_mode != None:
            manifest.vm_safe_mode = vm_safe_mode  # 修改安全模式
        process_list = manifest.get_process_list()
        authorities = manifest.package_name + '.authorities'
        print('Add provider %s' % provider_name)
        manifest.add_provider(provider_name, authorities)  # 主进程
        for i, process in enumerate(process_list):
            sub_provider_name = provider_name + '$InnerClass' + str(i + 1)
            print('Add provider %s in process %s' %
                  (sub_provider_name, process))
            manifest.add_provider(
                sub_provider_name, authorities + str(i), process)

        if activity_list:
            for activity in activity_list:
                print('Add activity %s' % activity['name'])
                manifest.add_activity(
                    activity['name'], activity['exported'], activity['process'])

        # 合并dex文件
        print('Merge dex %s' % merge_dex_path)
        classes_dex_path = tempfile.mktemp('.dex')

        dex_index = 1
        low_memory = False
        while True:
            dex_file = 'classes%s.dex' % (dex_index if dex_index > 1 else '')
            if not apk_file.get_file(dex_file):
                # 作为最后一个classes.dex
                with open(merge_dex_path, 'rb') as f:
                    apk_file.add_file(dex_file, f.read())
                print('Save dex %s to %s' % (merge_dex_path, dex_file))
                break

            if force_append or low_memory:
                # 低内存下直接跳过已有dex，进行加速
                dex_index += 1
                continue

            if os.path.exists(classes_dex_path):
                os.remove(classes_dex_path)
            apk_file.extract_file(dex_file, classes_dex_path)
            try:
                merge_dex(classes_dex_path, [
                          classes_dex_path, merge_dex_path], max_heap_size)
            except TooManyMethodsError:
                print('Merge dex into %s failed due to methods number' % dex_file)
                dex_index += 1
            except OutOfMemoryError:
                print('Merge dex into %s failed due to out of memory error' % dex_file)
                low_memory = True
                dex_index += 1
            else:
                print('Merge dex into %s success' % dex_file)
                with open(classes_dex_path, 'rb') as f:
                    apk_file.add_file(dex_file, f.read())
                break

        if dex_index > 1:
            # 合并进非主dex只支持5.0以上系统
            print('WARNING: APK can only be installed in android above 5.0')
            manifest.min_sdk_version = 21

        manifest.save()

        if res_file_list:
            for src_path, dst_path in res_file_list:
                with open(src_path, 'rb') as f:
                    print('Copy file %s => %s' % (src_path, dst_path))
                    data = f.read()
                    apk_file.add_file(dst_path, data)

        for it in apk_file.list_dir('META-INF'):
            if it.lower().endswith('.rsa'):
                print('Signature file is %s' % it)
                tmp_rsa_path = tempfile.mktemp('.rsa')
                apk_file.extract_file('META-INF/%s' % it, tmp_rsa_path)
                orig_signature = get_apk_signature(tmp_rsa_path).strip()
                os.remove(tmp_rsa_path)
                logging.info('%s signature is %s' %
                             (manifest.package_name, orig_signature))
                signature_dict[manifest.package_name] = orig_signature
                break
        else:
            raise RuntimeError('Can not find .sf file in META-INF dir')

        for it in apk_file.list_dir('META-INF'):
            apk_file.delete_file('META-INF/%s' % it)

    out_apk_list = []

    # 写入原始签名信息
    print('Write original signatures: %s' % json.dumps(signature_dict))
    temp_dir = tempfile.mkdtemp('-repack')
    for i, apk_file in enumerate(apk_file_list):
        apk_file.add_file('assets/qt4a_package_signatures.txt',
                          json.dumps(signature_dict))
        file_name = os.path.split(apk_path_or_list[i])[-1][:-4] + '-repack.apk'
        tmp_apk_path = os.path.join(temp_dir, file_name)
        apk_file.save(tmp_apk_path)
        new_path = resign_apk(tmp_apk_path)
        os.remove(tmp_apk_path)
        out_apk_list.append(new_path)
    if len(out_apk_list) == 1:
        return out_apk_list[0]
    else:
        return out_apk_list


if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''qpath模块单元测试
'''

try:
    from unittest import mock
except:
    import mock
import unittest

from qt4a.qpath import QPath

class TestQPath(unittest.TestCase):
    '''QPath类测试用例
    '''
    
    def test_id(self):
        self.assertEqual(QPath('/Id="webview"').parsed_qpath, [{'Id': ['=', 'webview']}])
    
    def test_type(self):
        self.assertEqual(QPath('/Type="WebView"').parsed_qpath, [{'Type': ['=', 'WebView']}])
    
    def test_text(self):
        self.assertEqual(QPath('/Text="标题"').parsed_qpath, [{'Text': ['=', '标题']}])
    
    def test_desc(self):
        self.assertEqual(QPath('/Desc="文本"').parsed_qpath, [{'Desc': ['=', '文本']}])
        
    def test_instance(self):
        self.assertEqual(QPath('/Text="标题" /Instance=1').parsed_qpath, [{'Text': ['=', '标题']}, {'Instance': ['=', 1]}])
    
    def test_maxdepth(self):
        self.assertEqual(QPath('/Type="TextView" /Text="消息" && MaxDepth=3').parsed_qpath, [{'Type': ['=', 'TextView']}, {'MaxDepth': ['=', 3], 'Text': ['=', '消息']}])
        
if __name__ == '__main__':
    unittest.main()


# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''APK安装包
'''

import zipfile
from io import BytesIO

class APKFile(object):
    '''
    '''
    def __init__(self, apk_path):
        self._apk_path = apk_path
        self._dir_tree = {}  # 目录树
        with open(self._apk_path, 'rb') as f:
            apk_data = f.read()
            self._fp = zipfile.ZipFile(BytesIO(apk_data), mode='r')
            for it in self._fp.filelist:
                self._dir_tree[it] = None
    
    def _get_item(self, rav_path):
        '''get ZipInfo item
        '''
        for it in self._dir_tree:
            filename = it
            if isinstance(it, zipfile.ZipInfo): filename = it.filename
            if filename == rav_path:
                return it
        return None
            
    def get_file(self, rav_path):
        '''获取apk内部文件的句柄
        
        :param rav_path: 文件在apk内的相对路径
        :type  rav_path: string
        '''
        it = self._get_item(rav_path)
        if it != None:
            data = self._fp.read(it.filename)
            return BytesIO(data)
        return None
    
    def list_dir(self, dir_rav_path):
        '''列举安装包中的文件列表
        
        :param dir_rav_path: 安装包中的目录相对路径
        :type dir_rav_path:  string
        '''
        result = []
        if dir_rav_path[-1] != '/': dir_rav_path += '/'
        for it in self._dir_tree:
            if isinstance(it, zipfile.ZipInfo): it = it.filename
            if it.startswith(dir_rav_path):
                result.append(it.split('/')[-1])
        return result
    
    def add_file(self, rav_path, file_data):
        '''添加文件
        '''
        it = self._get_item(rav_path)
        if not it: it = rav_path
        self._dir_tree[it] = file_data
        
    def delete_file(self, rav_path):
        '''删除安装包内的文件
        
        :param rav_path: 文件在apk内的相对路径
        :type  rav_path: string
        '''
        it = self._get_item(rav_path)
        if it != None: 
            del self._dir_tree[it]
    
    def extract_file(self, rav_path, save_path):
        '''提取文件到本地
        
        :param rav_path: 文件在apk内的相对路径
        :type rav_path:  string
        :param save_path:保存路径
        :type save_path: string
        '''
        it = self._get_item(rav_path)
        if not it: raise RuntimeError('file %s not in apk %s' % (rav_path, self._apk_path))
        data = self._fp.read(rav_path)
        with open(save_path, 'wb') as f:
            f.write(data)
        
    def save(self, save_path):
        '''保存
        '''
        out_fp = zipfile.ZipFile(save_path, 'w', zipfile.ZIP_DEFLATED)
        for it in self._dir_tree:
            data = self._dir_tree[it]
            if not data: 
                filename = it
                if isinstance(it, zipfile.ZipInfo): filename = it.filename
                data = self._fp.read(filename)
            out_fp.writestr(it, data)
        out_fp.close()
    
if __name__ == '__main__':
    pass
    



# -*- coding:UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''管理命令
'''

from __future__ import print_function

import argparse
import logging
import os
import sys
import time

try:
    raw_input
except:
    raw_input = input
    
def install_qt4a_driver(args):
    from qt4a.androiddriver.adb import ADB
    from qt4a.androiddriver.androiddriver import copy_android_driver
    device_id = None
    device_list = ADB.list_device()
    if args.serialno and args.serialno not in device_list:
        raise RuntimeError('Device %s not found' % args.serialno)
    elif args.serialno:
        device_id = args.serialno
    else:
        if len(device_list) == 0:
            raise RuntimeError('No Android device found')
        elif len(device_list) == 1:
            device_id = device_list[0]
        elif len(device_list) > 1:
            text = '\nCurrent Android device list:\n'
            for i, dev in enumerate(device_list):
                text += '%d. %s\n' % ((i + 1), dev)

            while True:
                print(text)
                result = raw_input('Please input the index of device to install driver:\n')
                if result.isdigit():
                    if int(result) > len(device_list):
                        sys.stderr.write('\nIndex %s out of range\nValid index range: [1, %d]\n' % (result, len(device_list)))
                        time.sleep(0.1)
                        continue
                    device_id = device_list[int(result) - 1]
                else:
                    if not result in device_list:
                        sys.stderr.write('\nDevice %r not exist\n' % result)
                        time.sleep(0.1)
                        continue
                    device_id = result
                break
        
    print('Device "%s" will install driver...' % device_id)
    copy_android_driver(device_id, args.force)
    print('Install QT4A driver to %s completely.' % device_id)


def qt4a_repack_apk(apk_path_or_list, debuggable=True, max_heap_size=0, force_append=False):
    '''重打包apk
    
    :param apk_path_or_list: apk路径或apk路径列表
    :type  apk_path_or_list: string/list
    :param debuggable: 重打包后的apk是否是调试版本：
                           True - 是
                           False - 否
                           None - 与原apk保持一致
    :type  debuggable: bool/None
    :param max_heap_size: 能够使用的最大堆空间，单位为：MB
    :type  max_heap_size: int/float
    '''
    from qt4a.apktool import repack
    cur_path = os.path.dirname(os.path.abspath(__file__))
    activity_list = [{
        'name': 'com.test.androidspy.inject.CmdExecuteActivity',
        'exported': True,
        'process': 'qt4a_cmd'
    }]

    # 添加QT4A测试桩文件
    file_path_list = []
    file_list = ['AndroidSpy.jar',
                'arm64-v8a/libdexloader.so',
                'arm64-v8a/libdexloader64.so',
                'arm64-v8a/libandroidhook.so',
                'armeabi/libdexloader.so',
                'armeabi/libandroidhook.so',
                'armeabi-v7a/libdexloader.so',
                'armeabi-v7a/libandroidhook.so',
                'x86/libdexloader.so',
                'x86/libandroidhook.so'
                ]
    tools_path = os.path.join(cur_path, 'androiddriver', 'tools')
    for it in file_list:
        file_path = os.path.join(tools_path, it)
        file_path_list.append((file_path, 'assets/qt4a/%s' % it))

    return repack.repack_apk(apk_path_or_list, 
        'com.test.androidspy.inject.DexLoaderContentProvider',
        os.path.join(cur_path, 'apktool', 'tools', 'dexloader.dex'),
        activity_list,
        file_path_list,
        debuggable,
        max_heap_size=max_heap_size,
        force_append=force_append
        )


def repack_apk(args):
    print('Repacking apk %s...' % (' '.join(args.path)))
    outpath = qt4a_repack_apk(args.path, args.debuggable, args.max_heap, force_append=args.force_append)
    print('Repack apk completely.\nOutput apk path is: ')
    if isinstance(outpath, list):
        for it in outpath:
            print(it)
    else:
        print(outpath)


def inspect_apk(args):
    from qt4a.apktool.manifest import AndroidManifest
    print('Apk %s info:' % args.path)
    apk = AndroidManifest(args.path)
    print('  Package name: %s' % apk.package_name)
    print('  Version: %s' % apk.version_name)
    print('  Minimun sdk: %s' % apk.min_sdk_version)
    print('  Targat sdk: %s' % apk.target_sdk_version)
    start_activity = apk.start_activity
    if start_activity.startswith('.'):
        start_activity = apk.package_name + start_activity
    print('  Start activity: %s' % start_activity)


def qt4a_manage_main():
    from qt4a.androiddriver.util import OutStream
    logging.root.level = logging.INFO
    sys.stdout = OutStream(sys.stdout)
    sys.stderr = OutStream(sys.stderr)
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help='subcommand')
    install_driver_parser = subparsers.add_parser('install-driver', help='install qt4a driver')
    install_driver_parser.add_argument('-s', '--serialno', help='device serialno')
    install_driver_parser.add_argument('-f', '--force', action='store_true', help='force install qt4a driver')
    install_driver_parser.set_defaults(func=install_qt4a_driver)
    
    repack_parser = subparsers.add_parser('repack-apk', help='repack apk file')
    repack_parser.add_argument('-p', '--path', nargs='*', required=True, help='path of apks to repack')
    repack_parser.add_argument('-d', '--debuggable', type=bool, default=True, help='whether apk debuggable after repack')
    repack_parser.add_argument('-m', '--max-heap', type=int, default=0, help='max heap size can use, unit is MB')
    repack_parser.add_argument('-a', '--force-append', action='store_true', default=False, help='force append the dex instead of merge')
    repack_parser.set_defaults(func=repack_apk)
    
    inspect_parser = subparsers.add_parser('inspect-apk', help='inspect apk file')
    inspect_parser.add_argument('-p', '--path', required=True, help='path of apk to inspect')
    inspect_parser.set_defaults(func=inspect_apk)

    args = parser.parse_args()

    if hasattr(args, 'func'): 
        args.func(args)
    else:
        parser.print_help()
        print('\n%s: error: too few arguments' % os.path.split(sys.argv[0])[-1], file=sys.stderr) # show error info in python3
        
        
if __name__ == '__main__':
    qt4a_manage_main()
    
    
    

